# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""define things that should be visible to the user."""

# pylint: disable=protected-access

from .accounts import *
from .track import *
from .lifetrack import *
from .backend import *
from .diff import *
from .locate import *
from .backends import *
from .version import *

__all__ = [
    'Track', 'Fences', 'Lifetrack', 'Locate', 'Directory', 'GPSIES', 'MMT', 'TrackMMT', 'Openrunner',
    'BackendDiff', 'WPTrackserver', 'Mailer', 'VERSION', 'Account', 'DirectoryAccount']


def prepare_backends():
    """Initialize the attribute "supported" for all backends."""
    for key in globals().keys():
        cls = globals()[key]
        if hasattr(cls, "__mro__") and cls is not Backend:
            if len(cls.__mro__) > 3 and cls.__mro__[-3] == Backend:
                cls._define_support()


prepare_backends()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines some helpers."""

import os
import datetime
import time
import logging
from math import isclose

from gpxpy.geo import length as gpx_length

__all__ = ['Duration', 'repr_timespan', 'uniq', 'remove_directory', 'is_track', 'collect_tracks',
           'positions_equal', 'pairs', 'add_speed', 'utc_to_local_delta']


class Duration:

    """A context manager showing time information for debugging."""

    # pylint: disable=too-few-public-methods

    def __init__(self, name):
        """See class docstring."""
        self.name = name
        self.start_time = datetime.datetime.now()

    def __enter__(self):
        """See class docstring.

        Returns:
            self

        """
        return self

    def __exit__(self, exc_type, exc_value, trback):
        """See class docstring."""
        logging.debug(
            '%s in %s %s-%s',
            datetime.datetime.now() - self.start_time,
            self.name, self.start_time, datetime.datetime.now())


def repr_timespan(start, end) ->str:
    """return a string representing the timespan.

    Returns:
        a string like #h#m

    """
    duration = end - start
    hours = duration.seconds // 3600
    minutes = (duration.seconds - hours * 3600) // 60
    hours += duration.days * 24
    return '{}:{:02}'.format(hours, minutes)


def uniq(lst):
    """return lst with unique elements."""
    seen = []
    for _ in lst:
        if _ not in seen:
            seen.append(_)
            yield _


def remove_directory(path):
    """If this fails, show directory content."""
    try:
        os.rmdir(path)
    except FileNotFoundError:
        logging.error("REMOVE_DIRECTORY %s: not found", path)
        raise
    except OSError as exc:
        logging.error('rmdir: errno: %s cannot remove directory: %s', exc, path)
        if os.path.exists(path):
            for _ in os.listdir(path):
                logging.error('  dir still has %s', _)


def is_track(value):
    """Return True or False without looking at the type, so we do not need to import Track."""
    return hasattr(value, 'id_in_backend')


def collect_tracks(sources):
    """A copied list with tracks combined from all sources, to be used in 'for'-loops.

    Returns:
        A list of tracks

    """
    if is_track(sources):
        return [sources]
    result = list()
    for source in sources:
        if is_track(source):
            result.append(source)
        else:
            logging.debug('')
            logging.debug('collecting tracks from %s %s', source.account.backend, source)
            result.extend(source)
    return result


def positions_equal(pos1, pos2, digits=4):
    """Check if both points have the same position.

    Args:
        digits: Number of after comma digits to compare

    Returns:
        True if so

    """
    _ = 1 / 10 ** digits
    return isclose(pos1.longitude, pos2.longitude, rel_tol=_) and isclose(pos1.latitude, pos2.latitude, rel_tol=_)


def add_speed(points, window: int = 2):
    """Add speed to points in m/sec.

    It uses the last X points for computation: distance and time
    between [current_point - window] an [current_point]

    Args:
        window: The number of last points to consider.

    """
    if not points:
        return
    points[0].gpxity_speed = 0.0
    start_idx = 0
    for _, point in enumerate(reversed(points)):  # noqa
        if hasattr(point, 'gpxity_speed'):
            start_idx = _
            break
    start_idx = len(points) - start_idx
    if start_idx == len(points):
        return

    target_points = points[start_idx:]
    for idx, target_point in enumerate(target_points):
        window_start = max(idx + start_idx - window, 0)
        start_point = points[window_start]
        window_distance = gpx_length(points[window_start:start_idx + idx])
        delta = (target_point.time - start_point.time)
        window_time = delta.days * 86400.0 + delta.seconds + delta.microseconds / 1000000.0
        if window_time:
            window_speed = window_distance / window_time
        else:
            window_speed = 0.0
        target_point.gpxity_speed = window_speed


def pairs(seq):
    """Return a list of all adjacent elements."""
    # pylint: disable=stop-iteration-return
    iterable = iter(seq)
    prev = next(iterable)
    for _ in iterable:
        yield prev, _
        prev = _


def utc_to_local_delta():
    """The difference local - utc.

    This is encapsulated because
    Returns: (timedelta) the difference.

    """
    return datetime.timedelta(seconds=time.localtime().tm_gmtoff)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This implements :class:`gpxity.wptrackserver.WPTrackserver`.

WPTrackserver talks directly to the WP mysql database holding the trackserver data.

This backend does not directly support Track.category, Track.status, Track.keywords.
All this is added to the description when writing and extracted when reading. So
the description in the backend will contain all keywords, but Track.description
does not.

The database has a maximum field length of 255 for strings. If it is exceeded, the
last part of the strings will be thrown away silently when writing into the database.

"""

# pylint: disable=protected-access

from gpxpy import gpx as mod_gpx

from ..backend import Backend
from ..gpx import Gpx
from ..util import add_speed, utc_to_local_delta

try:
    import MySQLdb
    HAVE_MYSQL = True
    import _mysql_exceptions
except ImportError:
    HAVE_MYSQL = False

__all__ = ['WPTrackserver']


class WPTrackserver(Backend):

    """Talk directly to the wordpress mysql database holding the trackserver data.

    See https://www.grendelman.net/wp/trackserver-wordpress-plugin/

    The section in the acounts file could look like::

        Account wp
            Backend WPTrackserver
            Username wordpress_username
            Url localhost
            Mysql wordpress_7@wordpress_7
            Password xxxx
            Fences 53.7505,10.7445/750

    Find the values for MySql in the wordpress config file (DB_USER and DB_NAME).

    Password is for the mysql user. Find it in the wordpress config file,
    look for DB_PASSWORD.

    """

    # pylint: disable=abstract-method

    test_is_expensive = False

    _keywords_marker = '\nKEYWORDS: '

    _max_length = {'title': 255, 'description': 255}

    def __init__(self, account=None):
        """See class docstring. The url is host."""
        super(WPTrackserver, self).__init__(account)
        self.__cached_db = None
        self.__cached_user_id = None

    def __connect_mysql(self):
        """Connect to the Mysql server.

        Returns: The db handle

        """
        try:
            user, database = self.account.mysql.split('@')
        except ValueError:
            raise Backend.BackendException('Url is illegal: {}'.format(self.url))
        try:
            result = MySQLdb.connect(
                host=self.url, user=user, passwd=self.account.password, database=database,
                autocommit=True, charset='utf8')
            if self.__cached_db:
                self.logger.info('reconnected to %s %s', self.url, database)
            else:
                self.logger.info('connected to %s %s', self.url, database)
            return result
        except _mysql_exceptions.Error as exc:
            raise Backend.BackendException(
                '{}: host={} user={} passwd={} database={}'.format(
                    exc, self.url, user, self.account.password, database))

    @property
    def _db(self):
        """Cached mysql handle.

        Returns: The handle.

        """
        if self.__cached_db is None:
            self.__cached_db = self.__connect_mysql()
        return self.__cached_db

    @property
    def _user_id(self):
        """Cached user_id from mysql.

        Returns: The user_id in wordpress

        """
        if self.__cached_user_id is None:
            if not self.account.username:
                raise self.BackendException('{} needs a username'.format(self.account))
            cursor = self.__exec_mysql('select id from wp_users where user_login=%s', [self.account.username])
            row = cursor.fetchone()
            if row is None:
                raise Backend.BackendException('WPTrackserver: User {} is not known'.format(self.account.username))
            self.__cached_user_id = row[0]
        return self.__cached_user_id

    def _encode_description(self, track):
        """Encode keywords in description.

        If description exceeds its maximum length, first remove keywords and then
        shorten the description.

        Returns: The string to be saved in the database.

        """
        def fmt_result():
            return '{}{}{}'.format(
                track.description, self._keywords_marker, ', '.join(kw_parts))

        max_length = self._max_length['description']
        kw_parts = track.gpx.keywords.split(', ')  # they already included encoded Category:  etc
        result = fmt_result()
        if len(result) > max_length:
            while kw_parts and len(result) > max_length:
                kw_parts = kw_parts[:-1]
                result = fmt_result()
        return result[:max_length]

    @classmethod
    def _decode_description(cls, gpx, value):
        """Extract keywords.

        Returns: The decoded description

        """
        if cls._keywords_marker in value:
            descr, raw_keywords = value.split(cls._keywords_marker)
            gpx.keywords = raw_keywords
        else:
            descr = value
        gpx.description = descr
        return descr

    def _gpx_from_headers(self, row):
        """Get header values out of row.

        Returns: Gpx()

        """
        result = Gpx()
        result.time = row[1]
        result.name = row[2]
        self._decode_description(result, row[3].replace('\r', ''))
        return result

    def _load_track_headers(self):
        """."""
        cmd = 'select id,created,name,comment,distance from wp_ts_tracks where user_id=%s'
        args = (self._user_id, )  # noqa
        cursor = self.__exec_mysql(cmd, args)
        for _ in cursor.fetchall():
            track = self._found_track(str(int(_[0])), self._gpx_from_headers(_))
            track.distance = _[4] / 1000.0

    @staticmethod
    def __point(row):
        """Make a GPX point.

        Returns:
            The point

        """
        time_delta = utc_to_local_delta()  # WPTrackserver wants local time
        return mod_gpx.GPXTrackPoint(
            latitude=float(row[0]),
            longitude=float(row[1]),
            time=row[2] - time_delta)

    def _read_all(self, track) ->None:
        """Read the full track."""
        assert track.id_in_backend
        cursor = self.__exec_mysql(
            'select latitude,longitude,occurred from wp_ts_locations where trip_id=%s', [track.id_in_backend])
        track.add_points([self.__point(x) for x in cursor.fetchall()])
        track.gpx.is_complete = True

    @staticmethod
    def __needs_insert(cursor, ident) -> bool:
        """Check if the header exists in the track table.

        Returns: True or False.

        """
        if ident is None:
            return True
        cursor.execute('select 1 from wp_ts_tracks where id=%s', (ident, ))  # noqa
        return len(cursor.fetchall()) == 0

    def _save_header(self, track):
        """Write all header fields. May set track.id_in_backend.

        Be aware that the track may still have 0 points (all fenced away).
        Returns: The new id_in_backend.

        """
        description = self._encode_description(track)
        title = track.title[:self._max_length['title']]
        # 1970-01-01 01:00:00 does not work. This is the local time but the minimal value 1970-01-01 ... is UTC
        track_time = track.first_time or Gpx.undefined_date
        track_distance = track.distance * 1000
        cursor = self._db.cursor()
        if self.__needs_insert(cursor, track.id_in_backend):
            if track.id_in_backend is None:
                cmd = 'insert into wp_ts_tracks(user_id,name,created,comment,distance,source)' \
                    ' values(%s,%s,%s,%s,%s,%s)'
                args = (self._user_id, title, track_time, description, track_distance, '')
                cursor = self.__exec_mysql(cmd, args)
                track.id_in_backend = str(int(cursor.lastrowid))
            else:
                self.__exec_mysql(
                    'insert into wp_ts_tracks(id,user_id,name,created,comment,distance,source) '
                    ' values(%s,%s,%s,%s,%s,%s,%s)',
                    (track.id_in_backend, self._user_id, title, track_time, description, track_distance, ''))
                self.logger.error('wptrackserver wrote missing header with id=%s', track.id_in_backend)
        else:
            self.__exec_mysql(
                'update wp_ts_tracks set name=%s,created=%s,comment=%s,distance=%s where id=%s',
                (title, track_time, description, track_distance, track.id_in_backend))
        return track.id_in_backend

    def _write_all(self, track) ->str:
        """save full gpx track.

        Since the file name uses title and title may have changed,
        compute new file name and remove the old files. We also adapt track.id_in_backend.

        Returns:
            the new track.id_in_backend

        """
        result = self._save_header(track)
        self.__exec_mysql('delete from wp_ts_locations where trip_id=%s', [int(result)])
        self.__write_points(track, track.points())
        return result

    def __write_points(self, track, points):
        """save points in the track."""
        points = list(points)
        time_delta = utc_to_local_delta()  # WPTrackserver wants local time
        cmd = 'insert into wp_ts_locations(trip_id, latitude, longitude, altitude,' \
            ' occurred, speed, comment, heading)' \
            ' values(%s, %s, %s, %s, %s, %s, "", 0.0)'
        args = [(
            track.id_in_backend, x.latitude, x.longitude, x.elevation or 0.0,
            x.time + time_delta, x.gpxity_speed if hasattr(x, 'gpxity_speed') else 0.0) for x in points]
        if args:
            self.__exec_mysql(cmd, args, many=True)

    def _remove_ident(self, ident: str) ->None:
        """backend dependent implementation."""
        cmd = 'delete from wp_ts_locations where trip_id=%s'
        self.__exec_mysql(cmd, [int(ident)])
        cmd = 'delete from wp_ts_tracks where id=%s'
        self.__exec_mysql(cmd, [int(ident)])

    def _change_ident(self, track, new_ident: str):
        """Change the id in the backend."""
        assert track.id_in_backend != new_ident
        if new_ident in self:
            raise ValueError(
                'New id_in_backend {} already exists in {}'.format(
                    new_ident, self.account))
        try:
            self.__exec_mysql(
                'update wp_ts_tracks set id=%s where id=%s',
                (new_ident, track.id_in_backend))
        except _mysql_exceptions.DataError as exc:
            raise ValueError(str(exc))
        self.__exec_mysql(
            'update wp_ts_locations set trip_id=%s where trip_id=%s',
            (new_ident, track.id_in_backend))

        self.logger.info('%s: renamed %s to %s', self.account, track.id_in_backend, new_ident)
        track.id_in_backend = new_ident

    @classmethod
    def is_disabled(cls) ->bool:
        """True if this backend is disabled by env variable GPXITY_DISABLE_BACKENDS.

        or because python mysql is not installed.

        This variable is a comma separated list of Backend class names.

        Returns:
            True if disabled

        """
        if not HAVE_MYSQL:
            return True
        return super(WPTrackserver, cls).is_disabled()

    def _lifetrack_start(self, track, points) ->str:
        """Start a new lifetrack with initial points.

        Returns:
            new_ident: New track id

        """
        new_ident = self._save_header(track)
        self._lifetrack_update(track, points)
        return new_ident

    def _lifetrack_update(self, track, points):
        """Update a lifetrack with points.

        Args:
            track: The lifetrack
            points: The new points

        """
        points = list(points)
        add_speed(list(track.points()), window=10)
        cmd = 'update wp_ts_tracks set distance=%s where id=%s'
        args = (track.distance * 1000, track.id_in_backend)
        self.__exec_mysql(cmd, args)
        self.__write_points(track, points)

    def __exec_mysql(self, cmd, args, many=False):
        """Wrapper.

        Returns:
            cursor or None if done nothing

        """
        def logit(prefix):
            if many:
                # only log the first one
                self.logger.debug(prefix + " executemany, first one: " + cmd, *args[0])
            else:
                self.logger.debug(prefix + ' ' + cmd, *args)

        def do_it():
            try:
                execute(cmd, args)
            except _mysql_exceptions.Error as exception:
                logit("MySQL Error: {} for".format(exception))
                raise
        cursor = self._db.cursor()
        execute = cursor.executemany if many else cursor.execute
        try:
            do_it()
        except _mysql_exceptions.OperationalError:
            # timeout disconnected
            self.__connect_mysql()
            do_it()
        return cursor

    @classmethod
    def _check_id_legal(cls, value):
        """Check if value is a legal id.

        If not, raise ValueError.

        """
        # it is not necessary to call BackendBase._check_id_legal
        if value is not None:
            if int(value) <= 0:
                # max is actually 2147483647 but the column is autoincrement
                # so just make mysql fail if we exceed that
                raise ValueError('{} not allowed as id_in_backend for WPTrackserver'.format(value))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.gpx`."""


from math import asin, sqrt, degrees
import datetime


# This code would speed up parsing GPX by about 30%. When doing
# that, GPX will only return str instead of datetime for times.
#
# import gpxpy.gpxfield as mod_gpxfield
# mod_gpxfield.TIME_TYPE=None

from gpxpy import gpx as mod_gpx
from gpxpy import parse as gpxpy_parse
from gpxpy.geo import length as gpx_length, Location
import geocoder
from geocoder.location import Location as Geocoder_location

from .util import repr_timespan, uniq, positions_equal

GPX = mod_gpx.GPX
GPXTrack = mod_gpx.GPXTrack
GPXTrackSegment = mod_gpx.GPXTrackSegment
GPXXMLSyntaxException = mod_gpx.GPXXMLSyntaxException

# see https://github.com/tkrajina/gpxpy/issues/150
Location.__hash__ = lambda x: int(x.latitude * 1000) + int(x.longitude * 1000) + int(x.elevation * 10)  # noqa


__all__ = ['Gpx']


class Gpx(GPX):

    """Wrapper around class GPX from gpxpy.

    Attributes:
        undefined_str_marker: A new Gpx() gets this value for all str attributes
            relevant for Gpxity. This lets :class:`~gpxity.track.Track` know if
            the wanted value is already known (by loading the list of tracks
            in the backend) or if the full track needs to be read.
        real_keywords: As decoded from keywords
        is_complete: False while we hold only metadata for the track.

        category: As decoded from keywords, no translation done
        public: Decoded from keywords
        ids: Decoded from keywords

    The user may either update keywords or keywords_only/category/public/ids.
    :meth:`encode` and :meth:`decode` synchronize them.

    Gpx.parse() sets both groups.

    """

    undefined_str = '__UXNXDXEXFXIXNXEXD__'
    undefined_date = datetime.datetime(year=1970, month=1, day=3, hour=1)

    def __init__(self):
        """Put 'undefined' markers into all fields of relevance for gpxity."""
        super(Gpx, self).__init__()
        self.name = Gpx.undefined_str
        self.description = Gpx.undefined_str
        self.keywords = Gpx.undefined_str
        self.time = Gpx.undefined_date

        self.real_keywords = list()
        self.category = Gpx.undefined_str
        self.public = Gpx.undefined_str
        self.ids = list()
        self.is_complete = False

    def encode(self):
        """Set keywords from real_keywords, category, public, ids."""
        all_kw = self.real_keywords[:]  # Make sure not to change the orignal
        if self.category != Gpx.undefined_str:
            all_kw.append('Category:{}'.format(self.category))
        all_kw.append('Status:{}'.format('public' if self.public is True else 'private'))
        for _ in self.ids:
            # TODO: encode the , to something else
            all_kw.append('Id:{}'.format(_))
        self.keywords = ', '.join(all_kw) or None

    def decode(self):
        """Extract real_keywords, category, public,ids from keywords."""
        if self.keywords is None:
            self.keywords = ''
        if self.keywords == Gpx.undefined_str:
            self.category = Gpx.undefined_str
            self.public = Gpx.undefined_str
            self.ids = list()
        else:
            data = [x.strip() for x in self.keywords.split(',')]
            if data == ['']:
                data = []
            real_keywords = list()
            ids = list()
            for keyword in data:
                _ = [x.strip() for x in keyword.split(':')]
                what = _[0]
                value = ':'.join(_[1:])
                if what == 'Category':
                    self.category = value
                elif what == 'Status':
                    self.public = value == 'public'
                elif what == 'Id':
                    ids.append(value)
                else:
                    real_keywords.append(keyword)
            self.ids = ids
            self.real_keywords = sorted(x[1:] if x.startswith('-') else x for x in real_keywords)

    @property
    def first_time(self) ->datetime.datetime:
        """datetime.datetime: start time of track.

        For a simpler implementation of backends, notably :class:`~gpxity.backends.mmt.MMT`
        we ignore gpx.time. Instead we return the time of the earliest track point.
        Only if there is no track point, return gpx.first_time. If that is unknown
        too, return None.

        For the same reason time is readonly.

        We assume that the first point comes first in time and the last
        point comes last in time. In other words, points should be ordered
        by their time.

        """
        try:
            return next(self.points()).time
        except StopIteration:
            return self.time

    @property
    def distance(self) ->float:
        """For me, the earth is flat.

        Returns:
            the distance in km, rounded to m. 0.0 if not computable.  # TODO: needs unittest

        """
        return round(gpx_length(list(self.points())) / 1000, 3)

    def add_points(self, points):
        """Just add points."""
        if points:
            if self.tracks:
                # make sure the same points are not added twice
                _ = self.tracks[-1].segments[-1]
                assert points != _.points[-len(points):]
            else:
                self.tracks.append(GPXTrack())
                self.tracks[0].segments.append(GPXTrackSegment())
            self.tracks[-1].segments[-1].points.extend(points)

    @classmethod
    def parse(cls, indata, is_complete: bool = True):
        """Parse xml data.

        Args:
            indata: may be a file descriptor or str
            is_complete: indata holds the entire track info, not just metadata

        Returns: Gpx()

        """
        result = Gpx()
        result.is_complete = is_complete
        if hasattr(indata, 'read'):
            indata = indata.read()
        if indata:
            # gpxpy.gpx has no classmethod constructor. This should be simpler.
            gpx = gpxpy_parse(indata)
            for _ in gpx.__slots__:
                setattr(result, _, getattr(gpx, _))
            result.workaround_for_gpxpy_issue_140()
            result.decode()
            if result.name is None:
                result.name = ''
            if result.description is None:
                result.description = ''
        return result

    def workaround_for_gpxpy_issue_140(self):
        """Remove empty track extension elements.

        Maybe we have to do that for other extension elements too.
        But I hope gpxity can soon enough depend on a fixed stable
        version of gpxpy.

        """
        for track in self.tracks:
            track.extensions = [x for x in track.extensions if len(x) or x.text is not None]

    def xml(self) ->str:
        """Produce exactly one line per trackpoint for easier editing (like removal of unwanted points).

        Returns: The xml string.

        """
        result = super(Gpx, self).to_xml()
        result = result.replace('</trkpt><', '</trkpt>\n<')
        result = result.replace('<copyright ></copyright>', '')   # gpxviewer does not accept such illegal xml
        result = result.replace('<link ></link>', '')
        result = result.replace('<author>\n</author>\n', '')
        result = result.replace('\n</trkpt>', '</trkpt>')
        result = result.replace('>\n<ele>', '><ele>')
        result = result.replace('>\n<time>', '><time>')
        result = result.replace('</ele>\n<time>', '</ele><time>')
        result = result.replace('.0</ele>', '</ele>')  # this could differ depending on the source
        # for newer gpxpy 1.3.3 which indents the xml:
        result = result.replace('\n      </trkpt>', '</trkpt>')
        result = result.replace('>\n        <ele>', '><ele>')
        result = result.replace('>\n        <time>', '><time>')
        result = result.replace('>\n        <name>', '><name>')
        result = result.replace('\n\n', '\n')
        if not result.endswith('\n'):
            result += '\n'
        return result

    @property
    def last_time(self) ->datetime.datetime:
        """The last time we received.

        Returns:

            The last time we received so far. If none, return None."""
        _ = self.last_point()
        return _.time if _ else None

    def speed(self) ->float:
        """Speed over the entire time in km/h or 0.0.

        Returns:
            The speed

        """
        time_range = (self.first_time, self.last_time)
        if time_range[0] is None or time_range[1] is None:
            return 0.0
        duration = time_range[1] - time_range[0]
        seconds = duration.days * 24 * 3600 + duration.seconds
        if seconds:
            return self.distance / seconds * 3600
        return 0.0

    def moving_speed(self) ->float:
        """Speed for time in motion in km/h.

        Returns:
            The moving speed

        """
        bounds = self.get_moving_data()
        if bounds.moving_time:
            return bounds.moving_distance / bounds.moving_time * 3.6
        return 0.0

    def __repr__(self) ->str:
        """The repr.

        Returns:
            the repr str

        """
        parts = []
        if self.keywords:
            parts.append(self.keywords)
        if self.name:
            parts.append(self.name)
        if self.first_time and self.last_time:
            parts.append(repr_timespan(self.first_time, self.last_time))
        elif self.first_time:
            parts.append(str(self.first_time))
        if self.distance:
            parts.append('{:4.2f}km'.format(self.distance))
        return '{}({})'.format(str(self), ' '.join(parts))

    def __str__(self) ->str:
        """The str.

        Returns: same as __repr__

        """
        return self.__repr__()

    def angle(self, precision=None) ->float:
        """For me, the earth is flat.

        Args:
            precision: After comma digits. Default is  6.

        Returns:
            the angle in degrees 0..360 between start and end.
            If we have no track, return 0

        """
        try:
            first_point = next(self.points())
        except StopIteration:
            return 0
        last_point = self.last_point()
        if precision is None:
            precision = 6
        delta_lat = round(first_point.latitude, precision) - round(last_point.latitude, precision)
        delta_long = round(first_point.longitude, precision) - round(last_point.longitude, precision)
        norm_lat = delta_lat / 90.0
        norm_long = delta_long / 180.0
        try:
            result = degrees(asin(norm_long / sqrt(norm_lat**2 + norm_long ** 2)))
            if norm_lat >= 0.0:
                result = (360.0 + result) % 360.0
            else:
                result = 180.0 - result
        except ZeroDivisionError:
            result = 0
        return result

    def segments(self):
        """
        A generator over all segments.

        Yields:
            GPXTrackSegment: all segments in all tracks

        """
        for track in self.tracks:
            for segment in track.segments:
                yield segment

    def points(self):
        """
        A generator over all points.

        Yields:
            GPXTrackPoint: all points in all tracks and segments

        """
        for segment in self.segments():
            for point in segment.points:
                yield point

    def point_list(self):
        """A flat list with all points.

        Returns:
            The list

        """
        return sum((x.points for x in self.segments()), [])

    def last_point(self):
        """Return the last point of the track. None if none."""
        # TODO: unittest for track without __gpx or without points
        try:
            return self.tracks[-1].segments[-1].points[-1]
        except IndexError:
            return None

    def adjust_time(self, delta):  # pylint: disable=arguments-differ
        """Add a timedelta to all times.

        gpxpy.gpx.adjust_time does the same but it ignores waypoints.
        A newer gpxpy.py has a new bool arg for adjust_time which
        also adjusts waypoints on request but I do not want to check versions.

        """
        super(Gpx, self).adjust_time(delta)
        for wpt in self.waypoints:
            wpt.time += delta
        self.time += delta

    def points_hash(self) -> float:
        """A hash that is hopefully different for every possible track.

        It is built using the combination of all points.

        Returns:
            The hash

        """
        result = 1.0
        for point in self.points():
            if point.longitude:
                result *= point.longitude
            if point.latitude:
                result *= point.latitude
            if point.elevation:
                result *= point.elevation
            _ = point.first_time
            if _:
                result *= (_.hour + 1)
                result *= (_.minute + 1)
                result *= (_.second + 1)
            result %= 1e20
        return result

    def points_equal(self, other, digits=4) ->bool:
        """
        Compare points for same position.

        Args:
            digits: Number of after comma digits to compare

        Returns:
            True if both tracks have identical points.

        All points of all tracks and segments are combined. Elevations are ignored.

        """
        # We do not use points_hash because we want to abort as soon as we know
        # they are different.
        if self.get_track_points_no() != other.get_track_points_no():
            return False
        for _, (point1, point2) in enumerate(zip(self.points(), other.points())):
            if not positions_equal(point1, point2, digits):
                return False
        return True

    def index(self, other, digits=4):
        """Check if this gpx contains other track gpx.

        This only works if all values for latitude and longitude are
        nearly identical.

        Useful if one of the gpx had geofencing applied.

        Args:
            digits: How many after point digits are used

        Returns:
            None or the starting index for other.points in self.points

        """
        self_points = self.point_list()
        other_points = other.point_list()
        for self_idx in range(len(self_points) - len(other_points) + 1):
            for other_idx, other_point in enumerate(other_points):
                if not positions_equal(self_points[self_idx + other_idx], other_point, digits):
                    break
            else:
                return self_idx
        return None

    @staticmethod
    def __time_diff(last_point, point):
        """Return difference in seconds, ignoring the date."""
        result = abs(last_point.hhmmss - point.hhmmss)
        if result > 33200:  # seconds in 12 hours
            result = 86400 - result
        return result

    @staticmethod
    def __point_is_near(last_point, point, delta_meter):
        """Return True if distance < delta_meter."""
        return abs(last_point.distance_2d(point)) < delta_meter

    def fix_jumps(self, minutes=30) ->bool:  # noqa pylint: disable=too-many-branches
        """Split segments at jumps.

        Whenever the time jumps back or more than X
        minutes into the future or the distance exceeds 5km,
        split the segment at that point.

        Returns: True if a split happened

        """
        result = False
        new_tracks = list()
        for track in self.tracks:
            new_segments = list()
            for segment in track.segments:
                if not segment.points:
                    result = True  # sort of - but also needs a rewrite
                    continue
                new_segment = GPXTrackSegment()
                new_segment.points.append(segment.points[0])
                for point in segment.points[1:]:
                    prev_point = new_segment.points[-1]
                    needs_break = False
                    if point.time is None and prev_point.time is not None:
                        needs_break = True
                    elif point.time is None and prev_point.time is None:
                        if point.distance_2d(prev_point) > 5000:
                            needs_break = True
                    elif point.time is not None and prev_point.time is None:
                        needs_break = True
                    elif point.time - prev_point.time > datetime.timedelta(minutes=minutes):
                        needs_break = True
                    elif point.time < prev_point.time:
                        needs_break = True
                    elif point.distance_2d(prev_point) > 5000:
                        needs_break = True
                    if needs_break:
                        result = True
                        new_segments.append(new_segment)
                        new_segment = GPXTrackSegment()
                    new_segment.points.append(point)
                new_segments.append(new_segment)
            new_track = GPXTrack()
            new_track.segments.extend(new_segments)
            new_tracks.append(new_track)
        if result:
            self.tracks = new_tracks
        return result

    def fix_orux(self) ->bool:
        """Try to fix Oruxmaps problems.

        1. the 24h bugs

        TODO: right now, result is always True even if nothing was done

        Returns: True if something changed.

        """
        all_points = list(uniq(self.points()))
        for _ in all_points:
            _.hhmmss = _.time.hour * 3600.0 + _.time.minute * 60
            _.hhmmss += _.time.second + _.time.microsecond / 1000000
        new_points = list([all_points.pop(0)])
        while all_points:
            last_point = new_points[-1]
            near_points = [x for x in all_points if self.__point_is_near(last_point, x, 10000)]
            if not near_points:
                near_points = all_points[:]
            nearest = min(near_points, key=lambda x: Gpx.__time_diff(last_point, x))
            new_points.append(nearest)
            all_points.remove(nearest)

        day_offset = 0
        point1 = None
        for point in new_points:
            if point1 is None:
                point1 = point
            else:
                point2 = point
                if point1.time - point2.time > datetime.timedelta(days=day_offset, hours=23):
                    day_offset += 1
                if day_offset:
                    point2.time += datetime.timedelta(days=day_offset)
                point1 = point2

        segment = GPXTrackSegment()
        segment.points.extend(new_points)

        self.tracks = list()
        self.tracks.append(GPXTrack())
        self.tracks[0].segments.append(segment)
        return True

    def time_offset(self, other):
        """If time and last_time have the same offset between both tracks, return that time difference.
        Otherwise return None."""
        def offset(point1, point2):
            """Returns the time delta if both points have a time."""
            if point1.time and point2.time:
                return point2.time - point1.time
            return None

        start_time_delta = offset(next(self.points()), next(other.points()))
        if start_time_delta:
            end_time_delta = offset(self.last_point(), other.last_point())
            if start_time_delta == end_time_delta:
                return start_time_delta
        return None

    def locate_point(self, point=0, segment=0, track=0, default_country=None):
        """Determine name of place for point.

        Saves that in point.name for caching.

        Args:
            point_idx: Index into point_list()

        Returns: tuple(point, located)
            located is True if locating was needed, False if we had it cached

        """
        point = self.tracks[track].segments[segment].points[point]
        result = not point.name
        if result:
            _ = Geocoder_location([point.latitude, point.longitude])
            place = geocoder.get(location=_, provider='osm', method='reverse')
            for _ in ('city', 'town', 'village', 'hamlet'):
                if hasattr(place, _):
                    name = getattr(place, _)
                    if name:
                        break
            if not name and 'address' in place.raw:
                name = place.raw['address'].get('city_district')
            if 'address' in place.raw:
                suburb = place.raw['address'].get('suburb')
                if suburb:
                    name += '-' + suburb
            if not name:
                name = place.address
            if place.country.lower() != default_country.lower():
                point.name = '{}, {}'.format(name, place.country)
            else:
                point.name = name
        return point.name, result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.lifetrack.Lifetrack`."""

# pylint: disable=protected-access

import datetime

from .track import Track

__all__ = ['Lifetrack']


class LifetrackTarget:

    """A single target of a lifetracking instance."""

    def __init__(self, backend, use_id=None):
        """See class docstring."""
        self.backend = backend
        if use_id in backend:
            existing_track = backend[use_id]
            self.track = existing_track.clone()
            self.track.id_in_backend = use_id
        else:
            self.track = Track()
            self.track.id_in_backend = use_id
        self.started = False
        assert self.track.backend is None, 'TrackerTarget.track {} has backend {}'.format(
            self.track, self.track.backend)

    def update(self, points) ->str:
        """Update lifetrack into a specific track.

        Returns:
            the new id_in_backend if not yet started else None

        """
        if not points:
            if not self.started:
                raise Exception('Lifetrack needs initial points')
            else:
                raise Exception('Lifetrack.update needs points')
        new_ident = None
        points = self._prepare_points(points)
        self.track.add_points(points)
        if not self.started:
            # TODO: a unittest where points are empty here because of fences
            new_ident = self.backend._lifetrack_start(self.track, points)
            self.track.id_in_backend = new_ident
            self.started = True
        elif points:
            self.backend._lifetrack_update(self.track, points)
        assert self.track.id_in_backend
        assert self.track.backend is None, 'LifetrackTarget.track {} has backend {}'.format(
            self.track, self.track.backend)
        return new_ident

    def end(self):
        """End lifetracking for a specific backend."""
        if not self.started:
            raise Exception('Lifetrack not yet started')
        self.backend._lifetrack_end(self.track)

    def _prepare_points(self, points):
        """Round points and remove those within fences.

        Returns (list):
            The prepared points

        """
        result = [x for x in points if self.backend.account.fences.outside(x)]
        if len(result) < len(points):
            self.backend.logger.debug("Fences removed %d out of %d points", len(points) - len(result), len(points))
        self.track._round_points(result)
        return result


class Lifetrack:

    """Life tracking. The data will be forwarded to all given backends.

    Args:
        sender_ip: The IP of the client.
        target_backends (list): Those tracks will receive the lifetracking data.
        ids (list(str)): If given, use as id_in_backend. One for every backend.
            May be list(str) or str

    Attributes:
        done: Will be True after end() has been called.
        ids (str): The ids of all backends joined by '----'.

    """

    def __init__(self, sender_ip, target_backends, ids=None):
        """See class docstring."""
        assert sender_ip is not None
        if ids is None:
            ids = [None] * len(target_backends)
        elif isinstance(ids, str):
            ids = ids.split('----')
        self.sender_ip = sender_ip
        self.targets = [LifetrackTarget(target, use_id) for target, use_id in zip(target_backends, ids)]
        self.done = False

    def formatted_ids(self) ->str:
        """One string holding all backend ids.

        Returns: that string or None.

        """
        try:
            return '----'.join(x.track.id_in_backend for x in self.targets)
        except TypeError:
            return None

    def start(self, points, title=None, public=None, category=None):
        """Start lifetracking.

        Returns: The new id_in_backend

        """
        if title is None:
            title = str(datetime.datetime.now())[:16]
        if public is None:
            public = False

        for _ in self.targets:
            with _.track._decouple():
                # decouple because the _life* methods will put data into the backend
                _.track.title = title
                _.track.public = public
                _.track.category = category
                _.started = _.track.id_in_backend is not None
        self.update(points)
        return self.formatted_ids()

    def update(self, points):
        """Start or update lifetrack.

        If the backend does not support lifetrack, this just saves the track in the backend.

        Args:
            points(list): New points

        """
        for _ in self.targets:
            _.update(points)

    def end(self):
        """End lifetrack.

        If the backend does not support lifetrack, this does nothing."""
        for _ in self.targets:
            _.end()
        self.done = True

    def __str__(self):  # noqa
        return 'Lifetrack({}{})'.format(self.formatted_ids(), ' done' if self.done else '')

    def __repr__(self):  # noqa
        return str(self)



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""define things that should be visible to the user."""

from .directory import *
from .mmt import *
from .gpsies import *
from .trackmmt import *
from .mailer import *
from .wptrackserver import *
from .openrunner import *



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
implements test classes for Track.

They only use backend Directory, so there is no network traffic involved
(unless Directory is a network file system, of course).

"""

# pylint: disable=protected-access

import os
import sys
import io
import filecmp
import tempfile
import datetime
import random
from unittest import skipIf

from gpxpy import gpx as mod_gpx

from .basic import BasicTest, disabled
from ... import Track, Backend, Account, DirectoryAccount
from ...backend_base import BackendBase
from ...gpx import Gpx
from .. import Directory, MMT, GPSIES, Mailer, TrackMMT, WPTrackserver
from .. import Openrunner
from ...util import repr_timespan, positions_equal, remove_directory

# pylint: disable=attribute-defined-outside-init

GPXTrackPoint = mod_gpx.GPXTrackPoint


class TrackTests(BasicTest):

    """track tests."""

    @skipIf(*disabled(Directory))
    def test_init(self):
        """test initialisation."""
        track = Track()
        self.assertFalse(track.public)
        with self.temp_backend(Directory) as backend:
            track = Track()
            track._set_backend(backend)
            self.assertEqual(len(backend), 0)
            backend.add(track)
            self.assertEqual(len(backend), 1)

        with self.temp_backend(Directory, count=2) as backend:
            backend.add(Track())
            self.assertEqual(len(backend), 3)

        test_url = tempfile.mkdtemp(prefix=DirectoryAccount.prefix)
        self.assertTrue(os.path.exists(test_url))
        remove_directory(test_url)
        self.assertFalse(os.path.exists(test_url))
        try:
            with self.temp_backend(Directory, url=test_url):
                self.assertTrue(os.path.exists(test_url))
        finally:
            remove_directory(test_url)

    @skipIf(*disabled(Directory))
    def test_track_list(self):
        """test list of tracks."""
        with self.temp_backend(Directory) as directory:
            self.assertEqual(len(directory), 0)
            track1 = Track()
            directory.add(track1)
            self.assertIn(track1, directory)
            self.assertIsNotNone(track1.id_in_backend)
            track1.description = 'x'
            self.assertIsNotNone(track1.id_in_backend)

    def test_clone(self):
        """True if the clone is identical."""
        track1 = self.create_test_track()
        track2 = track1.clone()
        self.assertEqualTracks(track1, track2)
        count1 = track1.gpx.get_track_points_no()
        del track1.gpx.tracks[0].segments[0].points[0]
        self.assertEqual(count1, track1.gpx.get_track_points_no() + 1)
        self.assertNotEqualTracks(track1, track2)
        track2 = track1.clone()
        track2.gpx.tracks[-1].segments[-1].points[-1].latitude = 5
        self.assertNotEqualTracks(track1, track2)
        track2 = track1.clone()
        track2.gpx.tracks[-1].segments[-1].points[-1].longitude = 5
        self.assertNotEqual(track1, track2)
        track2 = track1.clone()
        last_point2 = track2.gpx.tracks[-1].segments[-1].points[-1]
        last_point2.elevation = 500000
        self.assertEqual(last_point2.elevation, 500000)
        # here assertNotEqualTracks is wrong because keys() are still identical
        self.assertTrue(track1.points_equal(track2))
        track1.gpx.tracks.clear()
        track2.gpx.tracks.clear()
        self.assertEqualTracks(track1, track2)

    def test_no_category(self):
        """category must return default value if not present in gpx.keywords."""
        category_default = Track.categories[0]
        track = Track()
        self.assertEqual(track.category, category_default)
        track.category = None
        self.assertEqual(track.category, category_default)
        with self.assertRaises(Exception):
            track.category = 'illegal value'
        self.assertEqual(track.category, category_default)
        with self.assertRaises(Exception):
            track.change_keywords('Category:illegal value')
        self.assertEqual(track.category, category_default)

    def test_duplicate_category(self):
        """try to add two categories to Track."""
        category_other = Track.categories[5]
        track = Track()
        track.category = category_other
        with self.assertRaises(Exception):
            track.change_keywords('Category:{}'.format(category_other))

    def test_remove_category(self):
        """remove category from Track."""
        category_default = Track.categories[0]
        category_other = Track.categories[5]
        track = Track()
        track.category = category_other
        self.assertEqual(track.category, category_other)
        track.category = None
        self.assertEqual(track.category, category_default)

    def test_no_public(self):
        """public must return False if not present in gpx.keywords."""
        track = Track()
        self.assertFalse(track.public)

    def test_duplicate_public(self):
        """try to set public via its property and additionally with change_keywords."""
        track = Track()
        track.public = True
        self.assertTrue(track.public)
        with self.assertRaises(Exception):
            track.change_keywords('Status:public')

    def test_remove_public(self):
        """remove and add public from Track using remove_keywords and change_keywords."""
        track = Track()
        track.public = True
        with self.assertRaises(Exception):
            track.change_keywords('-Status:public')
        self.assertTrue(track.public)
        with self.assertRaises(Exception):
            track.change_keywords('Status:public')
        self.assertTrue(track.public)

    def test_last_time(self):
        """Track.last_time."""
        track = self.create_test_track()
        gpx_last_time = track.gpx.tracks[-1].segments[-1].points[-1].time
        self.assertEqual(track.last_time, gpx_last_time)

    def test_one_line_per_trkpt(self):
        """One line per trackpoint."""
        track = self.create_test_track()
        xml = track.xml()
        self.assertNotIn('<link ></link>', xml)
        lines = xml.split('\n')
        self.logger.debug('xml is:%s', xml)
        start_lines = {x for x in lines if x.strip().startswith('<trkpt')}
        end_lines = {x for x in lines if x.strip().endswith('</trkpt>')}
        have_points = track.gpx.get_track_points_no()
        self.assertEqual(len(start_lines), have_points)
        self.assertEqual(len(end_lines), have_points)
        self.assertEqual(start_lines, end_lines)

    def test_parse(self):
        """check for Track parsing xml correctly."""
        track = self.create_test_track()
        track.keywords = ['Here are some keywords']
        xml = track.xml()
        gpx = Gpx.parse(xml)
        track2 = Track()
        track2.gpx = gpx
        self.assertEqualTracks(track, track2)
        self.assertEqual(track.keywords, track2.keywords)
        track2 = Track()
        track2.gpx = Gpx.parse(io.StringIO(xml))
        self.assertEqualTracks(track, track2)

    def test_combine(self):
        """combine values in track with newly parsed."""
        # Here, category is always from the domain Track.category, no backend involved.
        # first, does it overwrite?
        track = self.create_test_track()
        self.assertFalse(track.public)
        xml = track.xml()
        self.assertIn('Status:private', xml)
        track_category = track.category
        if track_category == 'Cycling':
            other_category = 'Running'
        else:
            other_category = 'Cycling'
        track2 = Track()
        track2.title = 'Title2'
        track2.description = 'Description2'
        track2.category = other_category
        track2.public = True
        track2.gpx = Gpx.parse(xml)
        self.assertEqual(track2.title, track.title)
        self.assertEqual(track2.description, track.description)
        self.assertEqual(track2.category, track.category)
        self.assertFalse(track2.public)
        self.assertEqual(track2.keywords, list())

        track.public = True
        xml = track2.xml()
        self.assertIn('Status:private', xml)
        track2 = Track()
        track2.category = Track.categories[3]
        self.assertEqual(track2.gpx.keywords, 'Category:{}, Status:private'.format(Track.categories[3]))
        track2.public = True
        self.assertEqual(track2.gpx.keywords, 'Category:{}, Status:public'.format(Track.categories[3]))
        track2.gpx = Gpx.parse(xml)
        self.assertFalse(track2.public)

        # second, does it keep old values if there are no new values?
        track = self.create_test_track()
        track.title = ''
        track.description = 'xx'
        xml = track.xml()
        if track.category == 'Cycling':
            other_category = 'Running'
        else:
            other_category = 'Cycling'

        track2 = Track()
        track2.title = 'Title2'
        track2.description = 'Description2'
        self.assertIn('<desc>Description2</desc>', track2.xml())
        track2.gpx = Gpx.parse(xml)
        self.assertEqual(track2.title, '')
        self.assertEqual(track2.description, 'xx')

    @skipIf(*disabled(Directory))
    def test_save_dir(self):
        """Correct files?."""
        with self.temp_backend(Directory) as directory:
            os.chmod(directory.url, 0o555)
            track = self.create_test_track()
            if os.getuid() == 0:
                # for root, this works even with 555
                directory.add(track)
                self.assertIsNotNone(track.backend)
            else:
                with self.assertRaises(OSError):
                    directory.add(track)
                self.assertIsNone(track.backend)
                os.chmod(directory.url, 0o755)
                directory.add(track)
            self.assertIsNotNone(track.backend)

    @skipIf(*disabled(Directory))
    def test_save(self):
        """save locally."""
        with self.temp_backend(Directory) as directory:
            dir2 = directory.clone()
            try:
                track = self.create_test_track()
                directory.add(track)
                self.assertEqual(len(directory), 1)
                aclone = track.clone()
                self.assertEqualTracks(track, aclone)

                self.assertEqual(len(dir2), 1)

                track2 = track.clone()
                self.assertEqualTracks(track, track2)
                directory.add(track2)
                self.assertEqual(len(directory), 2)
                dir2.add(track2)
                self.assertEqual(len(dir2), 2)

                track2_copy = dir2.add(track2.clone())
                self.assertEqualTracks(track, track2_copy)
                self.assertEqualTracks(track2, track2_copy)
                self.assertIs(track.backend, directory)
                self.assertIs(track2.backend, directory)
                self.assertIs(track2_copy.backend, dir2)
                self.assertEqual(len(directory), 2)
                self.assertEqual(len(dir2), 3)
                directory.scan()  # we changed it through dir2
                self.assertEqual(len(directory), 4)
                dir2.scan()
                self.assertEqual(len(directory), 4)
                title = 'whatevertitle'
                for _ in dir2:
                    _.id_in_backend = title
                trunk = os.path.join(directory.url, title)
                expected_names = [trunk + x + '.gpx' for x in ('.1', '.2', '.3', '')]
                files = sorted(
                    os.path.join(directory.url, x)
                    for x in os.listdir(directory.url) if x.endswith('.gpx'))
                self.assertEqual(files, expected_names)
                self.assertEqual(len(dir2), 4)
                directory.scan()
                dir2.merge(directory, remove=True)
                self.assertEqual(len(dir2), 1)
                filecmp.clear_cache()
            finally:
                dir2.detach()

    def test_add_points(self):
        """test Track.add_points."""
        point_count = 11
        track = Track()
        points = self._random_points(count=point_count)
        track.add_points(points)
        self.assertEqual(track.gpx.get_track_points_no(), point_count)
        with self.assertRaises(Exception):
            track.add_points(points)
        track.add_points(points[:-1])
        self.assertEqual(track.gpx.get_track_points_no(), point_count * 2 - 1)

    def test_points_equal(self):
        """test Track.points_equal."""
        for _ in range(100):
            points = self._random_points(count=7)
            track1 = Track()
            track1.add_points(points)
            track2 = track1.clone()
            points2 = list(track2.points())  # those are cloned points
            self.assertTrue(track1.points_equal(track2))
            track2.gpx.tracks.clear()
            track2.add_points(points2[:5])
            self.assertFalse(track1.points_equal(track2))
            track2.add_points(points2[5:])
            self.assertTrue(track1.points_equal(track2))

            old_long = track2.gpx.tracks[-1].segments[-1].points[-2].longitude
            track2.gpx.tracks[-1].segments[-1].points[-2].longitude += 1
            self.assertFalse(track1.points_equal(track2))
            track2.gpx.tracks[-1].segments[-1].points[-2].longitude = old_long
            self.assertTrue(track1.points_equal(track2))

            old_lat = track2.gpx.tracks[-1].segments[-1].points[-2].latitude
            track2.gpx.tracks[-1].segments[-1].points[-2].latitude += 1
            self.assertFalse(track1.points_equal(track2))
            track2.gpx.tracks[-1].segments[-1].points[-2].latitude = old_lat
            self.assertTrue(track1.points_equal(track2))

            track2.gpx.tracks[-1].segments[-1].points[-2].elevation += 1
            self.assertTrue(track1.points_equal(track2))
            track2.gpx.tracks[-1].segments[-1].points[-2].elevation -= 1

            old_long = track2.gpx.tracks[-1].segments[-1].points[-1].longitude
            track2.gpx.tracks[-1].segments[-1].points[-1].longitude += 1
            self.assertFalse(track1.points_equal(track2))
            a1_points = list(track1.points())
            a2_points = list(track2.points())
            a1_first = a1_points[0]
            a1_last = a1_points[-1]
            a2_first = a2_points[0]
            a2_last = a2_points[-1]
            self.assertNotEqual(
                track1.angle(), track2.angle(),
                'a1.first:{} a1.last:{} a2.first:{} a2.last:{}'.format(
                    a1_first, a1_last, a2_first, a2_last))
            track2.gpx.tracks[-1].segments[-1].points[-1].longitude = old_long
            self.assertTrue(track1.points_equal(track2))

    @skipIf(*disabled(Directory))
    def test_repr(self):
        """test __str__."""
        track = Track()
        self.assertNotIn('id:', str(track))
        with self.temp_backend(Directory) as directory:
            track = Track()
            track.title = 'Title'
            track.category = 'Running'
            track.add_points(self._random_points(10))
            first_distance = track.distance
            self.assertIn('public' if track.public else 'private', repr(track))
            self.assertIn('Running', repr(track))
            self.assertIn(repr_timespan(track.first_time, track.last_time), repr(track))
            self.assertTrue(repr(track).startswith(str(track)))
            self.assertTrue(repr(track).endswith(')'))
            track.add_points(self._random_points(count=5, root=track.last_point()))
            self.assertGreater(track.distance, first_distance)
            self.assertIn('km', repr(track))
            directory.add(track)

            # repr(track) must not fully load it
            clone = directory.clone()
            self.assertNotIn(' points', repr(clone[0]))
            self.assertEqual(clone[0].gpx.get_track_points_no(), 15)
            self.assertIn('km', repr(clone[0]))
            self.assertEqual(track.category, 'Running')
            self.assertEqual(clone[0].category, 'Running')

    def test_angle(self):
        """test Track.angle."""
        track1 = Track()
        track1.add_points(list())
        self.assertEqual(len(track1.gpx.tracks), 0)
        self.assertEqual(track1.angle(), 0)
        track1.add_points(self._random_points(1))
        del track1.gpx.tracks[0].segments[0]
        self.assertEqual(track1.angle(), 0)
        for _ in range(1000):
            track1 = Track()
            track1.add_points(self._random_points(2))
            angle = track1.angle()
            self.assertLess(angle, 360.001)
            self.assertGreater(angle, -0.001)

        track1 = Track()
        track1.add_points(self._random_points(2))
        first_point = None
        for point in track1.points():
            if first_point is None:
                first_point = point
            else:
                point.latitude = first_point.latitude
                point.longitude = first_point.longitude
        self.assertEqual(track1.angle(), 0)

    def test_key(self):
        """Track.key()."""
        title = 'This is a niße title'
        description = title + ' NOT - it is the description'
        category = Track.categories[3]
        public = True
        points = self._random_points(10)
        track = Track()
        track.title = title
        track.description = description
        track.category = category
        track.public = public
        track.add_points(points)
        key = track.key()
        self.assertIn('title:{}'.format(title), key)
        self.assertIn('description:{}'.format(description), key)
        self.assertIn('category:{}'.format(category), key)
        self.assertIn('public:True', key)
        self.assertIn('last_time:{}'.format(track.last_time), key)
        self.assertIn('angle:{}'.format(track.angle()), key)
        self.assertIn('points:{}'.format(track.gpx.get_track_points_no()), key)

    @skipIf(*disabled(Directory))
    def test_symlinks(self):
        """Directory symlinks."""
        with self.temp_backend(Directory) as directory:
            source = os.path.join(directory.url, 'deadlink')
            target = 'deadtarget'
            target_path = os.path.join(directory.url, target)
            with open(target_path, 'w') as target_file:
                target_file.write(' ')
            os.symlink('deadtarget', source)
            os.remove(target_path)
            directory.scan()  # this loads symlinks. It removes the dead link.
            self.assertFalse(os.path.exists(source))

    @skipIf(*disabled(Directory))
    def test_fs_encoding(self):
        """fs_encoding."""
        with self.temp_backend(Directory) as directory:
            track = Track()
            directory.add(track)
            org_ident = track.id_in_backend
            track.title = 'TITLE'
            self.assertEqual(track.id_in_backend, org_ident)
            self.assertEqual(track.title, 'TITLE')
            track.title = 'Tätel'
            self.assertEqual(track.title, 'Tätel')
            for title in ('a/b', '//', 'Ä/Ü', '...'):
                track.title = title
                self.assertEqual(track.title, title)
                self.assertNotEqual(track.id_in_backend, title)
                track.id_in_backend = track.title.replace('/', '_')
                self.assertEqual(track.id_in_backend, title.replace('/', '_'))

        prev_encoding = sys.getfilesystemencoding
        try:
            sys.getfilesystemencoding = lambda: 'wrong'
            with self.assertRaises(Backend.BackendException) as context:
                with self.temp_backend(Directory):
                    pass
            expect = (
                'Backend Directory needs a unicode file system encoding,'
                ' .* has wrong. Please change your locale settings.')
            self.assertRegex(str(context.exception), expect, msg='{} != {}'.format(context.exception, expect))
        finally:
            sys.getfilesystemencoding = prev_encoding

    def test_local_keywords(self):
        """Some keyword tests. More see in test_backends."""
        # Category: and Status: are special
        gpx = self._get_track_from_test_file('test').gpx
        gpx.keywords = 'Category:Cycling, Status:public'
        track = Track(gpx=gpx)
        self.assertEqual(track.keywords, list())

        # : is legal within a keyword
        gpx.keywords = 'Hello:Dolly'
        track = Track(gpx=gpx)
        self.assertEqual(track.keywords, list(['Hello:Dolly']))

        # keywords are sorted
        gpx.keywords = 'Hello,Dolly'
        track = Track(gpx=gpx)
        self.assertEqual(track.keywords, list(['Dolly', 'Hello']))

        # no comma within a keyword
        with self.assertRaises(Exception):
            track.change_keywords(['Bye,Sam'])

        # keywords as string
        track.change_keywords('Bye,Sam')
        self.assertEqual(track.keywords, ['Bye', 'Dolly', 'Hello', 'Sam'])

    def test_keyword_args(self):
        """'Track.keywords' must accept any variant of iterable."""
        track = Track()
        test_tracks = list(sorted(['a', self.unicode_string2]))
        track.keywords = set(test_tracks)
        self.assertEqual(track.keywords, test_tracks)
        track.keywords = reversed(test_tracks)
        self.assertEqual(track.keywords, test_tracks)
        track.change_keywords(test_tracks[0])
        self.assertEqual(track.keywords, test_tracks)
        track.keywords = test_tracks * 2
        self.assertEqual(track.keywords, test_tracks)

    @skipIf(*disabled(Directory))
    def test_in(self):
        """x in backend."""
        with self.temp_backend(Directory) as directory:
            track = Track()
            directory.add(track).id_in_backend = '56'
            self.assertEqual(track.id_in_backend, '56')
            self.assertIn(track, directory)
            self.assertIn(track.id_in_backend, directory)
            directory.remove_all()
            self.assertNotIn(track, directory)
            self.assertNotIn(track.id_in_backend, directory)

    @skipIf(*disabled(Directory))
    def test_getitem(self):
        """backend[idx]."""
        with self.temp_backend(Directory) as directory:
            directory.scan(now=True)
            track = Track()
            directory.add(track).id_in_backend = '56'
            self.assertIs(directory[0], track)
            self.assertIs(directory[track], track)
            self.assertIs(directory['56'], track)
            directory.remove_all()
            with self.assertRaises(IndexError):
                directory[0]  # pylint: disable=pointless-statement

    def test_adjust_time(self):
        """adjust_time()."""
        track = self.create_test_track()
        first_wp_time = track.gpx.waypoints[0].time
        first_trkpt_time = next(track.points()).time
        seconds10 = datetime.timedelta(seconds=10)
        track.adjust_time(seconds10)
        self.assertEqual(track.gpx.waypoints[0].time, first_wp_time + seconds10)
        self.assertEqual(next(track.points()).time, first_trkpt_time + seconds10)

    def test_overlapping_times(self):
        """Track.overlapping_times(tracks)."""
        now = datetime.datetime.now()
        track1 = self.create_test_track(start_time=now)
        seconds10 = datetime.timedelta(seconds=10)
        track2 = self.create_test_track(start_time=track1.last_time - seconds10)
        track3 = self.create_test_track(start_time=track1.last_time)
        self.assertEqual(track1.last_time - seconds10, track2.first_time)
        group1 = list([track1, track2, track3])
        track4 = self.create_test_track(start_time=track3.last_time + seconds10)
        group2 = list([track4, track4])
        self.assertEqual(list(Track.overlapping_times(group1 + group2)), list([group1, group2]))
        group2 = list([track4])
        self.assertEqual(list(Track.overlapping_times(group1 + group2)), list([group1]))

    @skipIf(*disabled(Directory))
    def test_header_changes(self):
        """Change fields loaded by track scan, before _load_full() is done."""
        with self.temp_backend(Directory, count=1) as backend:
            backend2 = backend.clone()
            backend2[0].description = 'test'
            self.assertTrackFileContains(backend2[0], '<trk>')
        with self.temp_backend(Directory, count=1) as backend:
            backend2 = Directory(DirectoryAccount(backend.url))
            backend2[0].title = 'test title'
            self.assertTrackFileContains(backend2[0], '<trk>')
        with self.temp_backend(Directory, count=1) as backend:
            backend2 = Directory(DirectoryAccount(backend.url))
            backend2[0].category = backend2.supported_categories[2]
            self.assertTrackFileContains(backend2[0], '<trk>')

    @skipIf(*disabled(Directory))
    def test_remove_track(self):
        """If a backend has several identical tracks, make sure we remove the right one."""
        with self.temp_backend(Directory, count=1) as backend:
            track = backend[0]
            track_id = track.id_in_backend
            track2 = track.clone()
            backend.add(track2)
            backend.remove(track2)
            self.assertEqual(backend[0].id_in_backend, track_id)

    def test_header_data(self):
        """Test usage of Track._header_data."""
        # TODO: still needed?
        track = Track()
        gpx_track = self.create_test_track()
        track.distance = 5000
        self.assertEqual(track.distance, 5000)
        track.gpx = Gpx.parse(gpx_track.xml())
        self.assertEqual(track.distance, gpx_track.distance)

    @skipIf(*disabled(WPTrackserver))
    def test_merge_track(self):
        """Check if everything is correctly merged."""
        track1 = self.create_test_track()
        track1.title = '44432321'
        track1.keywords = 'KeyA,KeyB,KeyA'
        track1.ids = ['wptrackserver_unittest:5', '/tmp/x.gpx']
        track2 = track1.clone()
        track2.title = 'Track2-title'
        track2.ids = ['wptrackserver_unittest:5', 'wptrackserver_unittest:6', 'tmp/y.gpx']
        msg = track1.merge(track2, partial_tracks=True)
        for _ in msg:
            self.logger.debug(_)
        self.assertEqual(track1.gpx.get_track_points_no(), track2.gpx.get_track_points_no())
        self.assertTrue(track1.points_equal(track2, digits=9))
        self.assertEqual(track1.title, 'Track2-title')
        self.assertEqual(
            track1.ids,
            ['wptrackserver_unittest:5', '/tmp/x.gpx', 'wptrackserver_unittest:6', 'tmp/y.gpx'])

    def test_merge_partial_tracks(self):
        """Test Track.merge(partial_tracks=True)."""

        track1 = self.create_test_track()
        track1.title = '44432321'
        track1.keywords = 'KeyA,KeyB,KeyA'
        track2 = track1.clone()
        track2.title = 'Track2-title'
        self.assertTrue(track1.points_equal(track2, digits=9))

        track2.add_points(self._random_points(5, root=track1.last_point()))
        msg = track1.merge(track2, partial_tracks=True)
        for _ in msg:
            self.logger.debug(_)
        self.assertEqual(track1.gpx.get_track_points_no(), track2.gpx.get_track_points_no())
        self.assertTrue(track1.points_equal(track2, digits=9))
        self.assertEqual(track1.title, 'Track2-title')

        points2 = track2.point_list()
        points2[2].latitude = 5
        with self.assertRaises(Exception) as context:
            msg = track1.merge(track2, partial_tracks=True)
        self.assertEqual(
            str(context.exception),
            'Cannot merge {} with 27 points into {} with 27 points'.format(track2, track1))

    def test_all_backend_classes(self):
        """Test Backend.all_backend_classes."""
        all_classes = [x.__name__ for x in Backend.all_backend_classes()]
        expected = [Directory, GPSIES, MMT, Mailer, Openrunner, TrackMMT, WPTrackserver]
        expected = [x.__name__ for x in expected if not x.is_disabled()]
        self.assertEqual(all_classes, expected)

    def parse_objectnames(self, cases):
        """Helper for test_parse_objectname."""
        for string, expect_account_str, expect_backend, expect_ident in cases:
            account, track_id = BackendBase.parse_objectname(string)
            self.assertEqual(
                str(account), expect_account_str, 'str(account) wrong in test case:{}'.format(string))
            self.assertEqual(
                account.backend, expect_backend, 'backend wrong in test case:{}'.format(string))
            self.assertEqual(
                track_id, expect_ident, 'track_id wrong in test case:{}'.format(string))

    @skipIf(*disabled(Directory))
    def test_parse_objectname_directory(self):
        """Test Backend.parse_objectname for directory."""
        save = os.getenv('HOME'), os.getcwd()
        try:
            prefix = DirectoryAccount.prefix
            abs_prefix = os.path.abspath(prefix)
            os.chdir(prefix)
            test_home = os.path.abspath('subdir')
            os.environ['HOME'] = test_home  # for ~ in pathname
            cases = (
                ('.', '', 'Directory', None),
                ('subdir', 'subdir/', 'Directory', None),
                ('abc', '', 'Directory', 'abc'),
                ('subdir/abc', 'subdir/', 'Directory', 'abc'),
                ('subdir/sub2', 'subdir/sub2/', 'Directory', None),
                ('subdir/sub2/sub3/xy', 'subdir/sub2/sub3/', 'Directory', 'xy'),
                ('~/sub2', os.path.join(abs_prefix, 'subdir/sub2/'), 'Directory', None),
                ('~/sub2/sub3/xy', os.path.join(abs_prefix, 'subdir/sub2/sub3/'), 'Directory', 'xy'),
                ('missing_dir/24', 'missing_dir/', 'Directory', '24'),
                ('wptrackserver_unittest', 'wptrackserver_unittest/', 'Directory', None),
                ('wptrackserver_unittest/24', 'wptrackserver_unittest/', 'Directory', '24'),
                (os.path.join(test_home, 'sub2/sub3/xy'), os.path.join(test_home, 'sub2/sub3/'), 'Directory', 'xy'),
            )

            subdirs = list()
            subdirs.append(os.path.join(prefix, 'subdir'))
            subdirs.append(os.path.join(subdirs[0], 'sub2'))
            subdirs.append(os.path.join(subdirs[1], 'sub3'))
            subdirs.append(os.path.join(prefix, 'wptrackserver_unittest'))
            try:
                for _ in subdirs:
                    os.mkdir(_)
                self.parse_objectnames(cases)
            finally:
                for _ in reversed(subdirs):
                    remove_directory(_)
        finally:
            os.environ['HOME'] = save[0]
            os.chdir(save[1])

    def test_parse_objectname_other(self):
        """Test Backend.parse_objectname for other than Directory."""
        for cls in Backend.all_backend_classes():
            if cls is Directory:
                continue
            acc_name = cls.__name__.lower() + '_unittest'
            cases = (
                (acc_name + ':', acc_name + ':', cls.__name__, None),
                (acc_name + ':24', acc_name + ':', cls.__name__, '24'),
            )
            self.parse_objectnames(cases)
            break

    def test_fences(self):
        """Test fences."""

        # TODO: check accounts parsing

        for illegal in (
                '', 'a/b', '5.4.3/3.0/10', '5.4.3/3/10', '5/6/7/8'
        ):
            with self.assertRaises(ValueError, msg='fence "{}" is illegal'.format(illegal)):
                account = Account(fences=illegal)

        points = set(self._random_points())
        with self.temp_backend(Directory) as directory:
            accounts = (
                directory.account,
                DirectoryAccount('.', fences=None),
                Account(fences=' '.join("{}/{}/{}".format(
                    x.latitude, x.longitude, 500) for x in random.sample(points, 3))),
            )
            for account in accounts:
                fences = account.fences
                inside = {x for x in points if not fences.outside(x)}
                outside = {x for x in points if fences.outside(x)}
                self.assertEqual(inside | outside, points)
                self.assertEqual(len(inside & outside), 0)
                for point in inside:
                    self.assertFalse(fences.outside(point))
                for point in outside:
                    self.assertTrue(fences.outside(point))

    def test_openrunner_point_encoding(self):
        """Test Openrunner encoding/decoding of points."""
        for track, result in [
                ([(50.0, 7.0), (60.0, 8.0)], True),
                ([(-50.1, -7.2), (0.1, 8.4)], True),
                ([(-50.12, -7.23), (0.12, 8.45)], True),
                ([(-50.124, -7.234), (0.125, 8.458)], True),
                ([(-50.1041, -7.2354), (0.1325, 8.7458)], True),
                ([(-50.10341, -7.23554), (0.13325, 8.7458)], True),
                ([(-50.109341, -7.203554), (0.133425, 8.74258)], False),
        ]:
            points = [GPXTrackPoint(latitude=lat, longitude=lon) for lat, lon in track]
            enc_dec = Openrunner._decode_points(Openrunner._encode_points(points))
            self.assertEqual(result, all(positions_equal(*x, digits=10) for x in zip(points, enc_dec)), track)  # noqa



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.backend.Backend`."""

# pylint: disable=protected-access

import datetime
from inspect import getmembers, isfunction
from contextlib import contextmanager
import logging
from copy import deepcopy

from .accounts import Account
from .track import Track
from .util import collect_tracks
from .gpx import Gpx

from .backend_base import BackendBase
__all__ = ['Backend']


class Backend(BackendBase):

    """A place where tracks live. Something like the filesystem or http://mapmytracks.com.

    A Backend should hold only tracks for one person, and they
    should not overlap in time. This is not enforced but sometimes
    behaviour is undefined if you ignore this.

    A Backend be used as a context manager.

    A Backend allows indexing by normal int index, by :class:`Track <gpxity.track.Track>`
    and by :attr:`Track.id_in_backend <gpxity.track.Track.id_in_backend>`.
    :literal:`if 'ident' in backend` is possible.
    len(backend) shows the number of tracks. Please note that Code
    like :literal:`if backend:` may not behave as expected. This will be False if the backend
    has no track. If that is not what you want, consider :literal:`if backend is not None`

    The backend will automatically synchronize. So something like :literal:`len(Backend())` will work.
    However, some other Backend pointing to the same storage or even a different process
    might change things. If you want to cope with that, use :meth:`scan`.

    Not all backends support all methods. The unsupported methods
    will raise NotImplementedError. As a convenience every backend
    has a list **supported** to be used like :literal:`if 'track' in backend.supported:`
    where `track` is the name of the method.

    Backends support no locking. If others modify a backend concurrently, you may
    get surprises. It is up to you to handle those.

    Some backends may use cookies.

    Args:
        account (:class:`~gpxity.accounts.Account`): The account to be used.
            Alternatively a dict can be passed to build an ad hoc :class:`~gpxity.accounts.Account`
            instance.

    Attributes:
        supported (set(str)): The names of supported methods. Creating the first instance of
            the backend initializes this. Only methods which may not be supported are mentioned here.
            If a particular value write_* like write_public does not exist, the entire track is written instead
            which normally results in a new ident for the track.

            Some special values are:
                * rename: allows assigning values to id_in_backend

        full_support (set(str)): All possible values for the supported attribute.
        url (str): the address. May be a real URL or a directory, depending on the backend implementation.
            Every implementation may define its own default for url. Must never end with '/' except for
            Directory(url='/').
        test_is_expensive: For internal use. If True, the self tests will reduce test cases and try to
            avoid too much usage of the backend.
        max_field_sizes: Some backends have a limited size for some attributes like keywords. This
            is only an approximative guess. The backend will not protect you from overriding it
            but the unittests will try to stay within those limits.
        point_precision: The precision supported by this backend. We are never more precise than 6.
            That is the digits after the decimal separator.
        supported_categories: The categories supported by this backend. The first one is used as default.

    """

    # pylint: disable=too-many-instance-attributes

    class NoMatch(Exception):
        """Is raised if a track is expected to pass the match filter but does not"""

    supported = set()

    default_url = None  # Override in the backends

    test_is_expensive = True

    max_field_sizes = {}

    _category_decoding = dict()
    _category_encoding = dict()

    full_support = (
        'scan', 'remove', 'write', 'write_title', 'write_public',
        'own_categories', 'rename',
        'write_category', 'write_description', 'keywords', 'write_add_keywords', 'write_remove_keywords')

    _max_length = dict()

    point_precision = 5

    _timeout = None

    __all_backends = dict()

    # It is important that we have only one global session
    # because gpsies.com seems to have several servers and their
    # synchronization is sometimes slower than expected. See
    # cookie "SERVERID".
    _session = dict()

    def __init__(self, account):
        """See class docstring."""
        if self.is_disabled():
            raise Backend.BackendException('class {} is disabled'.format(self.__class__.__name__))
        if not isinstance(account, Account):
            raise Exception('Backend() wants an Account')
        account.config = deepcopy(account.config)
        self.account = account
        if 'url' not in self.account.config:
            self.account.config['url'] = self.default_url
        if 'backend' not in self.account.config:
            self.account.config['backend'] = self.__class__.__name__
        self._decoupled = False
        self.__tracks = list()
        self._tracks_fully_listed = False
        self.__match = None
        self.logger = logging.getLogger(str(self))
        # do not want to see "Resetting dropped connection"
        logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        self._cached_subscription = None  # to be used by specific Backend classes

    @property
    def timeout(self):
        """Timeout from account or class default.

        Returns: The timeout

        """
        return self.account.timeout or self._timeout

    @property
    def url(self):
        """get self.account.url.

        This also makes sure Backend.url is not writable.

        Returns: The url

        """
        if self.account.url is None:
            self.account.config['url'] = self.default_url
        return self.account.url

    def _has_default_url(self) ->bool:
        """Check if the backend has the default url.

        Returns:
            True if so

        """
        if self.default_url is None:
            return False
        return self.url == self.default_url

    def _get_author(self) ->str:
        """Get the username for the account.

        Raise BackendException if no username is given.

        Returns:
            The username

        """
        author = self.account.username
        if not author:
            raise Backend.BackendException('{} needs a username'.format(self.url))
        return author

    def __str__(self) ->str:
        """A unique identifier for every physical backend.

        Two Backend() instances pointing to the same physical backend have the same identifier.

        Returns:
            A unique identifier

        """
        return self.account.name

    supported_categories = Track.categories

    @contextmanager
    def _decouple(self):
        """Context manager: disable automic synchronization for the backend.

        In that state, automatic writes of changes into
        the backend are disabled, and if you access attributes which
        would normally trigger a full load from the backend, they will not.
        Use this to avoid recursions.

        You should never need this unless you implement a new backend.

        """
        prev_decoupled = self._decoupled
        self._decoupled = True
        try:
            yield
        finally:
            self._decoupled = prev_decoupled

    @property
    def match(self):
        """Filter tracks.

        A function with one argument returning None or str. The backend will call
            this with every track and ignore tracks where match does not return None.
            The returned str should explain why the track does not match.

            If you change a track such that it does not match anymore, the exception
            NoMatch will be raised and the match stays unchanged.

        """
        return self.__match

    @match.setter
    def match(self, value):
        """see match.getter."""
        old_match = self.__match
        self.__match = value
        try:
            self.scan()
        except self.NoMatch:
            self.__match = old_match
            raise

    @classmethod
    def _define_support(cls):
        """If the first thing a method does is raising NotImplementedError, it is marked as unsupported."""
        support_mappings = {
            # map internal names to more user friendly ones. See doc for
            # Backend.supported.
            '_change_ident': 'rename',
            '_load_track_headers': 'scan',
            '_remove_ident': 'remove',
            '_write_all': 'write'}
        cls.supported = set()
        cls.supported.add('keywords')  # default
        if cls.supported_categories != Track.categories:
            cls.supported.add('own_categories')
        for name, method in getmembers(cls, isfunction):
            if name in support_mappings:
                if cls._is_implemented(method):
                    cls.supported.add(support_mappings[name])
            elif name.startswith('_write_') and name != '_write_attribute':
                if cls._is_implemented(method):
                    cls.supported.add(name[1:])

    @classmethod
    def decode_category(cls, value: str) ->str:
        """Translate the value from the backend into one out of Track.categories.

        Returns:
            The decoded name

        """
        if value in Track.categories:
            return value
        if value in Track._obsolete_categories:
            return Track._obsolete_categories[value]
        if value.capitalize() in Track.categories:
            return value.capitalize()
        if value not in cls._category_decoding:
            raise cls.BackendException('{} gave us an unknown track type "{}"'.format(cls.__name__, value))
        return cls._category_decoding[value]

    @classmethod
    def encode_category(cls, value: str) ->str:
        """Translate internal value (out of Track.categories) into the backend specific value.

        Returns:
            The encoded name

        """
        if value in Track._obsolete_categories:
            value = Track._obsolete_categories[value]
        if value in cls.supported_categories:
            return value
        if value.lower() in cls.supported_categories:
            return value.lower()
        if value in cls._category_encoding:
            return cls._category_encoding[value]
        for key, target in cls._category_decoding.items():
            if value.lower() == target.lower():
                return key
        if value == Gpx.undefined_str:
            return cls.encode_category(Track.categories[0])
        raise cls.BackendException('{} has no equivalent for "{}"'.format(cls.__name__, value))

    @staticmethod
    def _encode_keyword(value: str) ->str:
        """Replicate the translation the backend does. MMT for example capitalizes all words.

        Returns:
            the encoded keyword

        """
        return value

    @staticmethod
    def _encode_description(track) ->str:
        """A backend might put keywords into the description. WPTrackserver does.

        Returns: The string to be saved in the backend

        """
        return track.description

    @classmethod
    def _decode_description(cls, gpx, value):
        """A backend might put keywords into the description. WPTrackserver does.

        Returns: The description

        """
        gpx.description = value
        return value

    def get_time(self) ->datetime.datetime:  # pylint: disable=no-self-use
        """get time from the server where backend is located as a Linux timestamp.

        A backend implementation does not have to support this.

        Returns: datetime.datetime

        """
        return datetime.datetime.now()

    def _change_ident(self, track, new_ident: str):
        """Change the id in the backend.

        If new_ident already exists, the backend is free to
        change it to a unique name or to raise an Exception.

        """
        raise NotImplementedError

    def scan(self, now: bool = False) ->None:
        """Enforce a reload of the list of all tracks in the backend.

        This will be delayed until the list is actually needed again.

        If this finds an unsaved track not matching the current match
        function, an exception is thrown.
        Saved Tracks not matching the current match will no be loaded.

        Args:
            now: If True, do not delay scanning.

        """
        self._tracks_fully_listed = False
        if now:
            self._scan()

    @property
    def is_scanned(self) ->bool:
        """Check if the backend has already been scanned for tracks.

        Returns: (bool)
            The answer

        """
        return self._tracks_fully_listed

    def _scan(self) ->None:
        """load the list of all tracks in the backend if not yet done.
        Enforce this by calling :meth:`scan` first.
        """
        if not self._tracks_fully_listed and not self._decoupled:
            self._tracks_fully_listed = True
            unsaved = [x for x in self.__tracks if x.id_in_backend is None]
            if self.__match is not None:
                for track in unsaved:
                    self.matches(track, 'scan')
            self.__tracks = unsaved
            if 'scan' in self.supported:
                match_function = self.__match
                self.__match = None
                try:
                    # _load_track_headers loads ALL tracks, match will be
                    # applied in a second loop. This way the Backend implementations
                    # do not have to worry about the match code.
                    with self._decouple():
                        self._load_track_headers()
                finally:
                    self.__match = match_function
            if self.__match is not None:
                self.__tracks = [x for x in self.__tracks if self.matches(x)]

    def _found_track(self, ident: str, gpx=None):
        """Create an empty track for ident and inserts it into this backend.

        Returns:
            the new track

        # TODO: Do I still need the result? Try not to

        """
        result = Track(gpx)
        with self._decouple():
            result._set_backend(self)
            result.id_in_backend = ident
        self._append(result)
        return result

    def _load_track_headers(self):
        """Load all track headers and append them to the backend.

        The tracks will not be loaded if possible.

        """
        raise NotImplementedError()

    def _read_all_decoupled(self, track) ->None:
        """Decouple and call the backend specific _read_all."""
        with self._decouple():
            self._read_all(track)

    def _read_all(self, track) ->None:
        """fill the track with all its data from source."""
        raise NotImplementedError()

    def matches(self, track, exc_prefix: str = None):
        """match track against the current match function.

        Args:
            exc_prefix: If not None, use it for the beginning of an exception message.
                If None, never raise an exception

        Returns:
            True for match

        """
        if self.__match is None:
            return True
        match_error = self.__match(track)
        if match_error and exc_prefix:
            raise Backend.NoMatch('{}: {} does not match: {}'.format(exc_prefix, track, match_error))
        return match_error is None

    def _needs_full_save(self, changes) ->bool:
        """Do we have to rewrite the entire track?.

        Returns:
            True if we must save fully

        """
        for change in changes:
            if change == 'all':
                return True
            write_name = 'write_{}'.format(change.split(':')[0])
            if write_name not in self.supported:
                return True
        return False

    def add(self, track):
        """
        Add a track to this backend.

        No track already existing in this backend will be overwritten.
        If  :attr:`Track.id_in_backend <gpxity.track.Track.id_in_backend>`
        is not given, the backend will create a unique value. If it is given,
        the backend will raise an exception with the exception of
        :class:`~gpxity.backends.directory.Directory` which will append
        a number to make it unique.

        Note that some backends may reject a track if it is very
        similar to an existing track even if it belongs to some other user.

        If :attr:`Track.id_in_backend <gpxity.track.Track.id_in_backend>` is given,
        the Backend may use it or not. Some always assign it themselves like
        :class:`~gpxity.backends.mmt.MMT`.

        If the track object is already in the list of tracks, raise ValueError.

        If the track does not pass the current match function, raise an exception.

        Args:
            track (~gpxity.track.Track): The track we want to save in this backend.

        Returns:
            ~gpxity.track.Track: The saved track. If the original track lives in a different
            backend, a new track living in this backend will be created
            and returned.

        """
        if self._decoupled:
            raise Exception('A backend cannot save() while being decoupled. This is probably a bug in gpxity.')
        self.matches(track, 'add')
        if track.backend is not self and track.backend is not None:
            # we do not want clone() loading the track because
            # that cannot be done with fences applied
            track._load_full()
            with track.fenced(self.account.fences):
                new_track = track.clone()
        else:
            if any(x is track for x in self.__tracks):
                raise ValueError(
                    'Already in list: Track {} with id={}, have={}'.format(
                        track, id(track), ','.join(str(x) for x in self)))
            new_track = track
        try:
            with self._decouple():
                new_track._set_backend(self)
                self._write_all(new_track)
            self._append(new_track)
            track._clear_dirty()
            return new_track
        except Exception:
            # do not do self.remove. If we try to upload the same track to gpsies,
            # gpsies will assign the same trackid, and we come here. __tracks will
            # only hold the first uploaded track, and remove would remove that
            # instance instead of this one.
            # TODO: do we have a unittest for that case?
            self.__tracks = [x for x in self.__tracks if x is not new_track]
            with self._decouple():
                new_track.id_in_backend = None
                new_track._set_backend(None)
            raise

    def _new_ident(self, track) ->str:
        """Create an id for track.

        Returns:
            The new ident. If the backend does not
            create an ident in advance, return None. Such
            backends will return a new ident after writing.

        """

    def _rewrite(self, track, changes):
        """Rewrite the full track.

        Used only by Track when things change.

        """
        assert track.backend is self
        assert self._has_item(track.id_in_backend), '{}: its id_in_backend {} is not in {}'.format(
            track, track.id_in_backend, ' / '.join(str(x) for x in self))
        assert track._dirty

        needs_full_save = self._needs_full_save(changes)

        self.matches(track, '_rewrite')
        if needs_full_save:
            with track.fenced(self.account.fences):
                new_id = self._write_all(track)
            track.id_in_backend = new_id
        else:
            for change in changes:
                _ = change.split(self._dirty_separator)
                write_name = '_write_{}'.format(_[0])
                if len(_) == 1:
                    getattr(self, write_name)(track)
                elif len(_) == 2:
                    getattr(self, write_name)(track, _[1])
                else:
                    raise Exception('dirty {} got too many arguments:{}'.format(write_name, _[1:]))

    def _write_all(self, track) ->str:
        """the actual implementation for the concrete Backend.

        Writes the entire Track.

        Returns:
            The new id_in_backend

        """
        raise NotImplementedError()

    def remove(self, value) ->None:
        """
        Remove track. This can also be done for tracks not passing the current match function.

        Args:
            value: If it is not an :class:`~gpxity.track.Track`, :meth:`remove` looks
                it up by doing :literal:`self[value]`

        """
        track = value if hasattr(value, 'id_in_backend') else self[value]
        if track.id_in_backend:
            self._remove_ident(track.id_in_backend)
        with self._decouple():
            track._set_backend(None)
            try:
                self.__tracks = [x for x in self.__tracks if x.id_in_backend != track.id_in_backend]
            except ValueError:
                pass

    def _remove_ident(self, ident: str) ->None:
        """backend dependent implementation."""
        raise NotImplementedError()

    def _lifetrack_start(self, track, points) -> str:  # pylint: disable=unused-argument
        """Modelled after MapMyTracks. I hope this matches other services too.

        This will always produce a new track in the backend.

        Default is to just add the points to the track.

        Args:
            track(Track): Holds initial data and points already added.
                keep in mind that this process might restart which should
                be invisible to the target.
            points: Initial points

        Returns: The new id_in_backend

        For details see :meth:`Track.track() <gpxity.lifetrack.Lifetrack.start>`.

        """
        if track.id_in_backend not in self:
            track.id_in_backend = self.add(track.clone()).id_in_backend
        return track.id_in_backend

    def _lifetrack_update(self, track, points):
        """If the backend does not support lifetrack, just add the points to the track.

        Args:
            track(Track): Holds initial data
            points: If None, stop tracking. Otherwise, start tracking
                and add points.

        For details see :meth:`Track.track() <gpxity.lifetrack.Lifetrack.update>`.

        """
        self[track.id_in_backend].add_points(points)

    def _lifetrack_end(self, track):
        """Default: Nothing needs to be done."""

    def remove_all(self):
        """Remove all tracks we know about.

        If their :attr:`id_in_backend` has meanwhile been changed
        through another backend instance or another process, we
        cannot find it anymore. We do **not** rescan all tracks in the backend.
        If you want to make sure it will be empty, call :meth:`scan` first.

        If you use a match function, only matching tracks will be removed."""
        for track in list(self):
            if self.matches(track):
                self.remove(track)

    def detach(self):
        """Should be called when access to the Backend is not needed anymore."""

    def __contains__(self, value) ->bool:
        """value is either an a track or a track id.

        Does NOT load tracks, only checks what is already known.

        Returns:
            True if we have the item

        """
        self._scan()
        return self._has_item(value)

    def _has_item(self, index) ->bool:
        """like __contains__ but for internal use: does not call _scan first.

        Must not call self._scan.

        Returns:
            True if we have the item

        """
        if hasattr(index, 'id_in_backend') and index in self.__tracks:
            return True
        if isinstance(index, str) and index in [x.id_in_backend for x in self.__tracks]:
            return True
        return False

    def __getitem__(self, index):
        """Allow accesses like alist[a_id].

        Do not call this when implementing a backend because this always calls scan() first.

        Instead use :meth:`_has_item`.

        Returns:
            the track

        """
        self._scan()
        if isinstance(index, int):
            return self.__tracks[index]
        for _ in self.__tracks:
            if _ is index or _.id_in_backend == index:
                return _
        raise IndexError

    def __len__(self) ->int:
        """do not call this when implementing a backend because this calls scan().

        Returns:
            the length

        """
        self._scan()
        return len(self.__tracks)

    def real_len(self) ->int:
        """len(backend) without calling scan() first.

        Returns:

            the length"""
        return len(self.__tracks)

    def _append(self, track):
        """Append a track to the cached list."""
        if track.id_in_backend is not None and not isinstance(track.id_in_backend, str):
            raise Exception('{}: id_in_backend must be str'.format(track))
        tracks_with_this_id = [x for x in self.__tracks if x.id_in_backend == track.id_in_backend]
        if tracks_with_this_id:
            assert len(tracks_with_this_id) == 1
            track_with_this_id = tracks_with_this_id[0]
            if track_with_this_id.backend is None:
                # we actually replace the unsaved track with the new one
                del self.__tracks[track_with_this_id]
        if track.id_in_backend is not None and any(x.id_in_backend == track.id_in_backend for x in self.__tracks):
            # cannot do "in self" because we are not decoupled, so that would call _scan()
            raise ValueError(
                'Backend.append(track {}): its id_in_backend {} is already in list: Track={}, list={}'.format(
                    str(track), track.id_in_backend, self[track.id_in_backend], self.__tracks))
        self.matches(track, 'append')
        self.__tracks.append(track)

    def __repr__(self):
        """do not call len(self) because that does things.

        Returns:
            The repr str

        """
        return '{}({} in {})'.format(self.__class__.__name__, len(self.__tracks), self.account)

    def __enter__(self):
        """See class docstring.

        Returns:
            self

        """
        return self

    def __exit__(self, exc_type, exc_value, trback):
        """See class docstring."""
        self.detach()

    def __iter__(self):
        """See class docstring.

        Returns:
            iterator over tracks

        """
        self._scan()
        return iter(self.__tracks)

    def __eq__(self, other) ->bool:  # TODO: use str
        """True if both backends have the same tracks.

        Returns:
            True if both backends have the same tracks

        """
        return {x.key() for x in self} == {x.key() for x in other}

    def __copy(self, other_tracks, remove, dry_run):
        """Copy other_tracks into self. Used only by self.merge().

        Returns:
            verbose messages

        """
        result = list()
        for old_track in other_tracks:
            if not dry_run:
                new_track = self.add(old_track)
            result.append('{} {} -> {}'.format(
                'blind move' if remove else 'blind copy', old_track, self.account if dry_run else new_track))
            if remove and not dry_run:
                old_track.remove()
        return result

    def __find_mergable_groups(self, tracks, partial_tracks: bool = False):
        """Find mergable groups.

        Returns:
            A list of tracks. The first one is the sink for the others

        """
        result = list()
        rest = list(self)
        rest.extend(x for x in tracks if str(x.backend) != str(self))
        while rest:
            root = rest[0]
            group = list([root])
            group.extend(x for x in rest[1:] if root.can_merge(x, partial_tracks)[0] is not None)  # noqa
            # merge target should be the longest track in self:
            group.sort(key=lambda x: (x.backend is self, x.gpx.get_track_points_no()), reverse=True)
            result.append(group)
            for _ in group:
                rest = [x for x in rest if x is not _]
        return result

    def merge(self, other, remove: bool = False, dry_run: bool = False, copy: bool = False,
              partial_tracks: bool = False) ->list:  # noqa
        """merge other backend or a single track into this one. Tracks within self are also merged.

        If two tracks have identical points, or-ify their other attributes.

        Args:
            other: The backend or a single track to be merged
            remove: If True, remove merged tracks
            dry_run: If True, do not really merge or remove
            copy: Do not try to find a matching track, just copy other into this Backend
            partial_tracks: If True, two tracks are mergeable if one of them contains the other one.

        Returns:
            list(str) A list of messages for verbose output

        """
        # pylint: disable=too-many-branches,too-many-locals
        # TODO: test for dry_run
        # TODO: test for merging single track
        # TODO: test for merging a backend or a track with itself. Where
        # they may be identical instantiations or not. For all backends.
        result = list()
        other_tracks = collect_tracks(other)
        if copy:
            return self.__copy(other_tracks, remove, dry_run)

        null_datetime = datetime.datetime(year=1, month=1, day=1)
        groups = self.__find_mergable_groups(other, partial_tracks)
        merge_groups = [x for x in groups if len(x) > 1]
        merge_groups.sort(key=lambda x: x[0].first_time or null_datetime)
        if merge_groups:
            result.append('{} mergable groups:'.format(len(merge_groups)))
            for _ in merge_groups:
                result.append('  {} ----> {}'.format(', '.join(str(x) for x in _[1:]), _[0]))  # noqa
        for destination, *sources in groups:
            if destination.backend is not self:
                new_destination = destination if dry_run else self.add(destination)
                result.append('{} {} -> {}'.format(
                    'move' if remove else 'copy', destination, self if dry_run else new_destination))
                if remove and not dry_run:
                    destination.remove()
                destination = new_destination
            for source in sources:
                result.extend(destination.merge(
                    source, remove=remove, dry_run=dry_run, partial_tracks=partial_tracks))
        return result

    @staticmethod
    def _html_encode(value) ->str:
        """encode str to something gpies.com accepts.

        Returns:
            the encoded value

        """
        if value is None:
            return ''
        return value.encode('ascii', 'xmlcharrefreplace').decode()

    def flush(self):
        """Some backends delay actual writing. This enforces writing.

        Currently, only the Mailer backend can delay, it will bundle all
        mailed tracks into one mail instead of sending separate mails
        for every track. Needed for lifetracking.

        """

    def clone(self):
        """return a clone."""
        return self.__class__(self.account)

    def _get_current_keywords(self, track):  # pylint:disable=no-self-use
        """A backend might be able to return the currently stored keywords.

        This is useful for unittests: Compare the internal state with what the
        backend actually says.

        """
        return track.keywords

    @classmethod
    def instantiate(cls, name: str):
        """Instantiate a Backend or a Track out of its identifier.

        Calls instantiate_backend and optionally changes result to Track.

        Args:
            name: The string identifier to be parsed

        Returns:
            A Track or a Backend. If the Backend has already been instantiated, return the cached value.
            If the wanted object does not exist, exception FileNotFoundError is raised.

        """
        result, track_id = cls.instantiate_backend(name)
        if track_id:
            try:
                result = result[track_id]
            except IndexError:
                raise FileNotFoundError('{} not found'.format(Track.identifier(result, track_id)))
        return result

    @classmethod
    def instantiate_backend(cls, name: str):
        """Instantiate a Backend.

        The full notation of an id_in_backend in a specific backend is
        similiar to what scp expects:

        Account:id_in_backend where Account is a reference to the accounts file.

        Locally reachable files or directories may be written without the leading
        Directory:. And a leading ~ is translated into the user home directory.
        The trailing .gpx can be omitted. It will be removed anyway for id_in_backend.

        If the file path of a local track (Directory) contains a ":", the file path
        must be absolute or relative (start with "/" or with "."), or the full notation
        with the leading Directory: is needed

        Args:
            name: The string identifier to be parsed

        Returns: tuple()
            * The first element is the Backend. If the Backend has already been instantiated, return the cached value.
                If the wanted object does not exist, exception FileNotFoundError is raised.
            * The second element is a track_id or None

        """
        account, track_id = cls.parse_objectname(name)
        cache_key = str(account)
        if cache_key in cls.__all_backends:
            result = cls.__all_backends[cache_key]
        else:
            result = cls.find_class(account.backend)(account)
            cls.__all_backends[cache_key] = result
        return result, track_id

    @property
    def subscription(self) ->str:
        """Get the subscription model. Like free, paid, plus, whatever.

        If the backend has no subscription model, return None.

        The unpaid subscription is granted to always return :literal:`free`.
        The most expensive subscription is granted to always return :literal:`full`.
        Intermediate values may vary.

        Because I  (the developer) have no paid account, I can test this only
        partially. Feedback is welcome!

        Returns: The name of the subscription or None

        """
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This implements :class:`gpxity.gpsies.GPSIES` for https://www.gpsies.com.

so ginge das mit dem API-Key: https://github.com/telemaxx/gpsiesreader/blob/master/gpsies3.py

"""

# pylint: disable=protected-access

from html.parser import HTMLParser
import datetime
import time
from collections import defaultdict

import requests

from .. import Backend
from ..gpx import Gpx

__all__ = ['GPSIES']


class GPSIESRawTrack:

    """raw data from the gpies html page."""

    # pylint: disable=too-few-public-methods
    def __init__(self):
        """See class docstring."""
        self.track_id = None
        self.title = None
        self.time = None
        self.distance = None
        self.public = True


class ParseGPSIESCategories(HTMLParser):  # pylint: disable=abstract-method

    """Parse the legal values for category from html."""

    def __init__(self):
        """See class docstring."""
        super(ParseGPSIESCategories, self).__init__()
        self.result = list(['biking'])

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        attributes = dict(attrs)
        if tag == 'input' and attributes['name'] == 'trackTypes':
            _ = attributes['id']
            if _ not in self.result:
                self.result.append(_)


class ParseGPIESEditPage(HTMLParser):  # pylint: disable=abstract-method

    """Parse the category value for a track from html."""

    def __init__(self):
        """See class docstring."""
        super(ParseGPIESEditPage, self).__init__()
        self.category = None

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        attributes = dict(attrs)
        if tag == 'input' and attributes['name'] == 'trackTypes' and 'checked' in attributes:
            self.category = attributes['id']


class ParseGPSIESList(HTMLParser):  # pylint: disable=abstract-method

    """get some attributes available only on the web page.

    Of course, this is highly unreliable. Just use what we can get."""

    def __init__(self):
        """See class docstring."""
        super(ParseGPSIESList, self).__init__()
        self.result = dict()
        self.result['tracks'] = list()
        self.track = None
        self.column = 0
        self.current_tag = None
        self.seeing_list = False
        self.after_list = False
        self.seeing_a = False
        self.seeing_warning = False

    def feed(self, data):
        """get data."""
        self.track = None
        self.column = 0
        self.current_tag = None
        self.seeing_list = False
        self.after_list = False
        self.seeing_a = False
        self.seeing_warning = False
        super(ParseGPSIESList, self).feed(data)

    def handle_starttag(self, tag, attrs):
        """starttag from the parser."""
        self.current_tag = tag
        attributes = defaultdict(str)
        for key, value in attrs:
            attributes[key] = value
        if tag == 'div' and 'alert-warning' in attributes['class']:
            self.seeing_warning = True
        if tag == 'tbody':
            self.seeing_list = True
        if not self.seeing_list:
            return
        if tag == 'tr':
            self.track = GPSIESRawTrack()
            self.column = 0
            self.seeing_a = False
        elif tag == 'td':
            self.column += 1
        elif self.after_list and tag == 'a':
            self.seeing_a = True
            value = attributes['value'].strip()
        elif tag == 'a' and 'href' in attributes and self.track.track_id is None:
            self.track.track_id = attributes['href'].split('fileId=')[1]
        elif tag == 'img' and self.track and 'lock.png' in attributes['src']:
            self.track.public = False

    def handle_endtag(self, tag):
        """handle end of track list."""
        if tag == 'tbody':
            self.seeing_list = False
            self.after_list = True

    def handle_data(self, data):
        """data from the parser."""
        data = data.strip()
        if not data:
            return
        if self.seeing_warning:
            raise GPSIES.BackendException(data)

        if self.seeing_list:
            if self.column == 3:
                if self.current_tag == 'i' and self.track.title is None:
                    self.track.title = data
            elif self.column == 4:
                if data.endswith('km'):
                    self.track.distance = float(data.replace(' km', '').replace(',', ''))
            elif self.column == 5:
                self.track.time = datetime.datetime.strptime(data, '%m/%d/%y')
                self.result['tracks'].append(self.track)


class GPSIES(Backend):

    """The implementation for gpsies.com.

    The track ident is the fileId given by gpsies.

    Searching arbitrary tracks is not supported. GPSIES only looks at the
    tracks of a specific user.

    GPSIES does not support keywords. If you upload a track with keywords,
    they will silently be ignored.

    Args:
        account (:class:`~gpxity.accounts.Account`): The account to be used.
            Alternatively a dict can be passed to build an ad hoc :class:`~gpxity.accounts.Account`
            instance.

    """

    # pylint: disable=abstract-method

    _default_description = 'None yet. Let everyone know how you got on.'

    supported_categories = (
        'biking', 'trekking', 'walking', 'jogging', 'climbing', 'racingbike', 'mountainbiking',
        'pedelec', 'skating', 'crossskating', 'handcycle', 'motorbiking', 'motocross', 'motorhome',
        'cabriolet', 'car', 'riding', 'coach', 'packAnimalTrekking', 'swimming', 'canoeing', 'sailing',
        'boating', 'motorboat', 'skiingNordic', 'skiingAlpine', 'skiingRandonnee', 'snowshoe', 'trailrunning',
        'speedhiking', 'wintersports', 'flying', 'train', 'sightseeing', 'geocaching', 'miscellaneous')

    _category_decoding = {
        'biking': 'Cycling',
        'boating': 'Rowing',
        'car': 'Driving',
        'climbing': 'Mountaineering',
        'handcycle': 'Cycling - Hand',
        'jogging': 'Running',
        'motocross': 'Enduro',
        'motorbiking': 'Motorcycling',
        'motorboat': 'Powerboating',
        'mountainbiking': 'Cycling - MTB',
        'packAnimalTrekking': 'Pack animal trekking',
        'racingbike': 'Cycling - Road',
        'riding': 'Horse riding',
        'skiingAlpine': 'Skiing - Alpine',
        'skiingNordic': 'Skiing - Nordic',
        'skiingRandonnee': 'Skiing - Touring',
        'snowshoe': 'Snowshoeing',
        'speedhiking': 'Hiking - Speed',
        'trailrunning': 'Running - Trail',
        'trekking': 'Hiking',
        'wintersports': 'Wintersports',
    }

    _category_encoding = {
        'Cycling - Foot': 'biking',
        'Cycling - Gravel': 'biking',
        'Cycling - Hand': 'handcycle',
        'Cycling - Indoor': 'biking',
        'Cycling - Road': 'racingbike',
        'Cycling - Touring': 'biking',
        'Driving': 'car',
        'Enduro': 'motocross',
        'Gliding': 'flying',
        'Hang gliding': 'flying',
        'Hiking': 'trekking',
        'Horse riding': 'riding',
        'Hot air ballooning': 'flying',
        'Jet skiing': 'motorboat',
        'Kayaking': 'boating',
        'Kiteboarding': 'sailing',
        'Longboard': 'miscellaneous',
        'Motor racing': 'motorbiking',
        'Motorcycling': 'motorbiking',
        'Mountaineering': 'climbing',
        'Nordic walking': 'walking',
        'Off road driving': 'car',
        'Orienteering': 'jogging',
        'Pack animal trekking': 'packAnimalTrekking',
        'Paragliding': 'flying',
        'Powerboating': 'motorboat',
        'River navigation': 'miscellaneous',
        'Rowing': 'boating',
        'Running': 'jogging',
        'Running - Road': 'jogging',
        'Running - Urban Trail': 'jogging',
        'Sea kayaking': 'boating',
        'Skateboarding': 'skating',
        'Skating - Inline': 'skating',
        'Skiing': 'wintersports',
        'Skiing - Alpine': 'skiingAlpine',
        'Skiing - Backcountry': 'skiingRandonnee',
        'Skiing - Crosscountry': 'skiingRandonnee',
        'Skiing - Roller': 'miscellaneous',
        'Skiing - Touring': 'skiingRandonnee',
        'Snowboarding': 'wintersports',
        'Snowshoeing': 'snowshoe',
        'Stand up paddle boarding': 'boating',
        'Swimrun': 'miscellaneous',
        'Wheelchair': 'miscellaneous',
        'Windsurfing': 'sailing',
        'Wintersports': 'wintersports',
    }

    default_url = 'https://www.gpsies.com'

    def __init__(self, account):
        """See class docstring."""
        super(GPSIES, self).__init__(account)
        self._session_response = None

    def _download_legal_categories(self):
        """Needed only for unittest.

        Returns: list(str)
            all legal values for category.

        """
        response = requests.post('{}?trackList.do'.format(self.url), timeout=self.timeout)
        category_parser = ParseGPSIESCategories()
        category_parser.feed(response.text)
        return sorted(category_parser.result)

    @property
    def session(self):
        """The requests.Session for this backend. Only initialized once.

        Returns:
            The session

        """
        ident = str(self)
        if ident not in self._session:
            author = self._get_author()
            if not self.account.password:
                raise self.BackendException('{}: Needs authentication data'.format(self.url))
            self._session[ident] = requests.Session()
            data = {'username': author, 'password': self.account.password}
            self._session_response = self._session[ident].post(
                '{}/loginLayer.do?language=en'.format(self.url),
                data=data, timeout=self.timeout)
            self._check_response(self._session_response)
            cookies = requests.utils.dict_from_cookiejar(self._session[ident].cookies)
            cookies['cookieconsent_dismissed'] = 'yes'
            self._session[ident].cookies = requests.utils.cookiejar_from_dict(cookies)
        return self._session[ident]

    def __post(self, action: str, data, files=None, track=None):
        """common code for a POST within the session.

        Returns:
            the response

        """
        for key in data:
            data[key] = self._html_encode(data[key])
        if data.get('fileDescription'):
            data['fileDescription'] = '<p>{}</p>'.format(data['fileDescription'])
        response = self.session.post(
            '{}/{}.do'.format(self.url, action), data=data, files=files, timeout=self.timeout)
        self._check_response(response, track)
        return response

    def _write_category(self, track):
        """change category on gpsies."""
        self._edit(track)

    def _write_description(self, track):
        """change description on gpsies."""
        self._edit(track)

    def _write_title(self, track):
        """change title on gpsies."""
        self._edit(track)

    def _write_public(self, track):
        """change public on gpsies."""
        self._edit(track)

    def _edit(self, track):
        """edit directly on gpsies."""
        assert track.id_in_backend
        data = {
            'edit': '',
            'fileDescription': track.description,
            'fileId': track.id_in_backend,
            'filename': track.title,
            'status': '1' if track.public else '3',
            'trackTypes': self.encode_category(track.category),
            'websiteUrl': ''}

        # in about 1 out of 10 cases this update does not work.
        # Doing that on the website with firefox shows the same problem.
        # So reload and compare until both are identical.
        copy = track.clone()
        copy.id_in_backend = track.id_in_backend
        ctr = 0
        while True:
            self.__post('editTrack', data, track=track)
            self._read_all(copy)
            if track.description != copy.description:
                msg = 'description: {} -> {}'.format(copy.description, track.description)
            elif track.title != copy.title:
                msg = 'title: {} -> {}'.format(copy.title, track.title)
            elif track.public != copy.public:
                msg = 'public: {} -> {}'.format(copy.public, track.public)
            elif self.encode_category(track.category) != self.encode_category(copy.category):
                msg = 'category: {}/{} -> {}/{}'.format(
                    copy.category, self.encode_category(copy.category),
                    track.category, self.encode_category(track.category))
            else:
                return
            ctr += 1
            time.sleep(1)
            if ctr > 50:
                raise Backend.BackendException(
                    'GPSIES: _edit fails to change track {}: {}'.format(track, msg))
            time.sleep(2)

    def _load_track_headers(self):
        """get all tracks for this user."""
    #    return
        response = self.__post('trackList', data={'username': self._get_author()})
        page_parser = ParseGPSIESList()
        page_parser.feed(response.text)
        hrefs = []
        for line in response.text.split('\n'):
            if 'pagination' in line:
                hrefs = [x for x in line.split(' ') if x.startswith('href')]
        for href in hrefs[2:-2]:
            href = href[1:-1]  # remove apostrophes
            parts = ''.join(href.split('?')[1:])
            parts = parts.split('&amp;')
            data = dict(x.split('=') for x in parts)  # noqa
            response = self.__post('userList', data=data)
            page_parser.feed(response.text)
        for raw_data in page_parser.result['tracks']:
            gpx = Gpx()
            gpx.name = raw_data.title
            gpx.time = raw_data.time
            public = raw_data.public
            if str(self) not in self._session:  # anonymous, no login
                public = True
            gpx.keywords = 'Status:{}, Category:{}'.format('public' if public else 'private', Gpx.undefined_str)
            track = self._found_track(raw_data.track_id, gpx)
            if raw_data.distance:
                track.distance = raw_data.distance

    def _read_category(self, track):
        """I found no way to download all attributes in one go."""
        data = {'fileId': track.id_in_backend}
        response = self.__post('editTrack', data, track=track)
        page_parser = ParseGPIESEditPage()
        page_parser.feed(response.text)
        track.category = self.decode_category(page_parser.category)

    def _read_all(self, track):
        """get the entire track. For gpies, we only need the gpx file."""
        data = {'fileId': track.id_in_backend, 'keepOriginalTimestamps': 'true'}
        response = self.__post('download', data=data, track=track)
        track.gpx = Gpx.parse(response.text)
        self._read_category(track)

    def _check_response(self, response, track=None):
        """are there error messages?."""
        trk_str = '{}: '.format(track) if track is not None else ''
        if response.status_code != 200:
            raise self.BackendException(response.text)
        if 'alert-danger' in response.text:
            _ = response.text.split('alert-danger">')[1].split('</div>')[0].strip()
            if '<li>' in _:
                _ = _.split('<li>')[1].split('</li>')[0]
            raise self.BackendException(trk_str + _)
        if 'alert-warning' in response.text:
            _ = response.text.split('alert-warning">')[1].split('<')[0].strip()
            ignore_messages = (
                'This track is deleted and only shown by a direct URL call.',
                'Track is not public, can be seen only by me',
                'GPSies is my hobby website and is funded by advertising'
            )
            if not any(x in _ for x in ignore_messages):
                self.logger.warning(trk_str + _)

    def _remove_ident(self, ident: str):
        """remove on the server."""
        data = {
            'delete': '',
            'fileDescription': 'n/a',
            'fileId': ident,
            'filename': 'n/a',
            'status': '1',
            'trackTypes': 'racingbike',
            'websiteUrl': ''}
        self.__post('editTrack', data=data)

    def _write_all(self, track) ->str:
        """save full gpx track on the GPSIES server.

        Returns:
            The new id_in_backend

        """
        files = {'formFile': (
            '{}.gpx'.format(self._html_encode(track.title)), track.xml(), 'application/gpx+xml')}
        data = {
            'fileDescription': track.description,
            'filename': track.title,
            'status': '1' if track.public else '3',
            'trackClassification': 'withoutClassification',
            'trackSimplification': '0',
            'trackTypes': self.encode_category(track.category),
            'uploadButton': ''}
        response = self.__post('upload', files=files, data=data, track=track)
        if 'Created' not in response.text:
            # not created
            raise self.BackendException('{}: {}'.format(track, response.text))
        new_ident = None
        for line in response.text.split('\n'):
            if 'fileId=' in line:
                new_ident = line.split('fileId=')[1].split('"')[0]
                break
        if not new_ident:
            raise self.BackendException('No fileId= found in response')
        if track.id_in_backend and track.id_in_backend != new_ident:
            self._remove_ident(track.id_in_backend)
        track.id_in_backend = new_ident
        return new_ident

    def detach(self):
        """also close session."""
        super(GPSIES, self).detach()
        if self.session:
            self.session.close()

    @classmethod
    def _define_support(cls):
        """GPSIES does not support keywords.

        TODO: encode them in the description.

        """
        super(GPSIES, cls)._define_support()
        cls.supported.remove('keywords')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This implements :class:`gpxity.directory.Directory`."""

# pylint: disable=protected-access

import os
import sys
import datetime
import tempfile

from collections import defaultdict

from gpxpy.gpx import GPXXMLSyntaxException

from .. import Backend, Track, DirectoryAccount
from ..util import remove_directory
from ..gpx import Gpx

__all__ = ['Directory', 'Backup']


class Backup:

    """A context manager making a backup of the gpx file.

    If an exception happened, restore the backup.
    Otherwise, remove it again.

    """

    # pylint: disable=too-few-public-methods

    def __init__(self, track):
        """See class docstring."""
        self.track = track
        self.old_id = track.id_in_backend
        self.old_pathname = None
        if self.old_id is not None:
            self.old_pathname = track.backend.gpx_path(self.old_id)
            if os.path.exists(self.old_pathname):
                os.rename(self.old_pathname, self.old_pathname + '.old')

    def __enter__(self):
        """See class docstring.

        Returns:
            self

        """
        return self

    def __exit__(self, exc_type, exc_value, trback):
        """See class docstring."""
        if exc_value:
            self.undo_rename()
            with self.track._decouple():
                self.track.id_in_backend = self.old_id
        else:
            if self.old_pathname is not None:
                if os.path.exists(self.old_pathname + '.old'):
                    os.remove(self.old_pathname + '.old')

    def undo_rename(self):
        """if something failed, undo change of file name and restore old file."""
        if self.old_pathname is not None:
            if os.path.exists(self.old_pathname):
                os.remove(self.old_pathname)
            if os.path.exists(self.old_pathname + '.old'):
                os.rename(self.old_pathname + '.old', self.old_pathname)


class Directory(Backend):

    """Uses a directory for storage.

    The filename minus the .gpx ending is used
    as :attr:`Track.id_in_backend <gpxity.track.Track.id_in_backend>`.

    If the :class:`~gpxity.directory.Directory` has a title but no id_in_backend,
    use the title as id_in_backend.
    Make the storage id unique by attaching a number if needed.
    A track without title gets a random name.

    The main directory (given by account.url) will have
    subdirectories YYYY/MM (year/month) with only the tracks for one month.
    Those are symbolic links to the main file and have the same file name.

    If :meth:`~gpxity.backend.Backend.save` is given a value for ident, this
    is used as id, the file name will be :literal:`id.gpx`.
    Otherwise, this backend uses :attr:`Track.title <gpxity.track.Track.title>` for the id.
    If a track has no title, it uses a random sequence of characters.
    Changing the title also changes the id.

    Args:
        Account: If its url is unset, this will create a temporary
            directory named :attr:`prefix`.X where X are some random characters.
            It will be removed in __exit__ / detach.
    Attributes:
        fs_encoding (str): The encoding for file system names. By default, we
            expect the file system being able to handle arbitrary UTF-8 encoded names
            except character '/' and special names '.' and '..'. If needed, we will introduce
            new possible values for fs_encoding like perhaps 'windows'. Gpxity will **never**
            support any other character set but UTF-8.
            Note that :attr:`fs_encoding` is independent of the platform we are running on - we
            might use a network file system.

    """

    # pylint: disable=abstract-method

    test_is_expensive = False

    def __init__(self, account):
        """See class docstring."""
        assert isinstance(account, DirectoryAccount)
        super(Directory, self).__init__(account)

        self.fs_encoding = sys.getfilesystemencoding()
        if not self.fs_encoding.lower().startswith('utf-8'):
            raise Backend.BackendException(
                'Backend Directory needs a unicode file system encoding, {} has {}.'
                ' Please change your locale settings.'.format(self, self.fs_encoding))

        self._symlinks = defaultdict(list)  # TODO: account.symlinks True
        self._load_symlinks()

    def __str__(self) ->str:
        """Used for formatting strings. Must be unique within the process.

        Returns:

            a unique identifier"""
        result = self.url
        if result:
            if result.startswith('./'):
                result = result[2:]
        else:
            result = '.'
        return result

    @staticmethod
    def _strip_gpx(name: str) ->str:
        """If it is there, strip traling .gpx.

        Returns:
            The stripped string.

        """
        if name.endswith('.gpx'):
            return name[:-4]
        return name

    def _load_symlinks(self, directory=None):
        """scan the subdirectories with the symlinks.

        If the content of a track changes, the symlinks might have to
        be adapted. But we do not know the name of the existing symlink anymore.

        So just scan them all and assign them to id_in_backend."""
        if directory is None:
            directory = self.url
        for dirpath, _, filenames in os.walk(directory):
            for filename in filenames:
                full_name = os.path.join(dirpath, filename)
                if os.path.islink(full_name):
                    if os.path.exists(full_name):
                        target = os.readlink(full_name)
                        gpx_target = self._strip_gpx(os.path.basename(target))
                        if full_name not in self._symlinks[gpx_target]:
                            self._symlinks[gpx_target].append(full_name)
                    else:
                        os.remove(full_name)

    def _new_id_from(self, ident_proposal: str) ->str:
        """Return not yet existant file name.

        Args:
            ident_proposal: If this proposal does not lead to a valid ident, create unique random ident.

        Returns:
            The new unique ident

        """
        value = self._sanitize_name(ident_proposal)
        if not value:
            value = os.path.basename(tempfile.NamedTemporaryFile(dir=self.url, prefix='').name)
        return self._make_ident_unique(value)

    @staticmethod
    def _make_path_unique(value) ->str:
        """If the file name already exists, apply a serial number.

        If value ends with .gpx, put the serial number in front of that.

        Returns:
            the unique path name

        """
        ctr = 0
        unique_value = value
        while os.path.exists(unique_value):
            ctr += 1
            if value.endswith('.gpx'):
                unique_value = '{}.{}.gpx'.format(value[:-4], ctr)
            else:
                unique_value = '{}.{}'.format(value, ctr)
        return unique_value

    def _make_ident_unique(self, value):
        """Return a unique ident."""
        path = Directory._make_path_unique(os.path.join(self.url, value + '.gpx'))
        return os.path.basename(path)[:-4]

    @staticmethod
    def _sanitize_name(value) ->str:
        """Change it to legal file name characters.

        Returns:
            the sanitized name

        """
        if value is None:
            return None
        return value.replace('/', '_')

    def gpx_path(self, ident) ->str:
        """The full path name for the local copy of a track.

        Returns:
            The full path name

        """
        assert isinstance(ident, str), '{} must be str'.format(ident)
        return os.path.join(self.url, '{}.gpx'.format(ident))

    def _list_gpx(self):
        """return a generator of all gpx files, with .gpx removed.

        Returns:
            A list of all gpx file names with .gpx removed

        """
        gpx_names = (x for x in os.listdir(self.url) if x.endswith('.gpx'))
        return (x.replace('.gpx', '') for x in gpx_names)

    @staticmethod
    def _get_field(data, name) ->str:
        """Get xml field out of data.

        Returns:
            The xml field

        """
        start_html = '<{}>'.format(name)
        end_html = '</{}>'.format(name)
        data = data.split(end_html)
        if len(data) > 1:
            data = data[0]
            data = data.split(start_html)
            if len(data) > 1:
                data = data[-1]
                if start_html not in data:
                    return data
        return None

    def _gpx_from_headers(self, ident):
        """Quick scan of file for getting some header fields.

        We do this by removing everything after the first point
        (or if no point is given everthin after metadata).

        Returns: Gpx

        """
        result = Gpx()
        with open(self.gpx_path(ident), encoding='utf8') as raw_file:
            data = raw_file.read(100000)
            head = None
            parts = data.split('</trkpt>')
            if len(parts) > 1:
                head = parts[0] + '</trkpt></trkseg></trk></gpx>'
            else:
                parts = data.split('</metadata>')
                if len(parts) > 1:
                    head = parts[0] + '</metadata></gpx>'
            if head:
                try:
                    result = Gpx.parse(head, is_complete=False)
                except GPXXMLSyntaxException:
                    self.logger.info(
                        '%s: Track metadata cannot be extracted, there is too much',
                        Track.identifier(self, ident))
        return result

    def _load_track_headers(self):
        """get all tracks for this user."""
        self._symlinks = defaultdict(list)
        self._load_symlinks()
        for _ in self._list_gpx():
            gpx = self._gpx_from_headers(_)
            self._found_track(_, gpx)

    def _read_all(self, track):
        """fill the track with all its data from source."""
        with open(self.gpx_path(track.id_in_backend), encoding='utf-8') as in_file:
            track.gpx = Gpx.parse(in_file.read())

    def _remove_symlinks(self, ident: str):
        """Remove its symlinks, empty symlink parent directories."""
        for symlink in self._symlinks[ident]:
            if os.path.exists(symlink):
                os.remove(symlink)
            symlink_dir = os.path.split(symlink)[0]
            try:
                os.removedirs(symlink_dir)
            except OSError:
                pass
        self._symlinks[ident] = list()

    def _remove_ident(self, ident: str):
        """Remove its symlinks and the file, in this order."""
        self._remove_symlinks(ident)
        gpx_file = self.gpx_path(ident)
        if os.path.exists(gpx_file):
            os.remove(gpx_file)

    def _symlink_path(self, track) ->str:
        """The path for the speaking symbolic link: YYYY/MM/title.gpx.

        Missing directories YYYY/MM are created.

        Returns:
            The path

        """
        ident = track.id_in_backend
        time = datetime.datetime.fromtimestamp(os.path.getmtime(self.gpx_path(ident)))
        by_month_dir = os.path.join(self.url, '{}'.format(time.year), '{:02}'.format(time.month))  # noqa
        if not os.path.exists(by_month_dir):
            os.makedirs(by_month_dir)
        else:
            # make sure there is no dead symlink with our wanted name.
            self._load_symlinks(by_month_dir)
        name = track.title or ident
        return self._make_path_unique(os.path.join(by_month_dir, self._sanitize_name(name)))

    def _new_ident(self, track):
        """Create an id for track.

        Returns: The new ident.

        """
        ident = track.id_in_backend
        if ident is None:
            if self.account.id_method == 'counter':
                try:
                    ident = str(max(int(x) for x in self._list_gpx()) + 1)
                except ValueError:
                    ident = '1'
            else:
                ident = self._new_id_from(None)
        return ident

    def _make_symlinks(self, track):
        """Make all symlinks for track."""
        ident = track.id_in_backend
        gpx_pathname = self.gpx_path(ident)
        link_name = self._symlink_path(track)
        basename = os.path.basename(gpx_pathname)
        link_target = os.path.join('..', '..', basename)
        os.symlink(link_target, link_name)
        if link_name not in self._symlinks[ident]:
            self._symlinks[ident].append(link_name)

    def _set_filetime(self, track):
        """Set the file modification time to track start time.
        If the track has no start time, do nothing."""
        time = track.first_time
        if time:
            _ = self.gpx_path(track.id_in_backend)
            os.utime(_, (time.timestamp(), time.timestamp()))

    def _change_ident(self, track, new_ident: str):
        """Change the id in the backend. Make it unique if needed."""
        assert track.id_in_backend != new_ident
        unique_id = self._new_id_from(new_ident)
        self._remove_symlinks(track.id_in_backend)
        self.logger.info('%s: renamed %s to %s', self.account, track.id_in_backend, unique_id)
        os.rename(self.gpx_path(track.id_in_backend), self.gpx_path(unique_id))
        track.id_in_backend = unique_id
        self._make_symlinks(track)

    def _write_all(self, track) ->str:
        """save full gpx track.

        Since the file name uses title and title may have changed,
        compute new file name and remove the old files. We also adapt track.id_in_backend.

        Returns:
            the new track.id_in_backend

        """
        new_ident = self._new_ident(track)

        with Backup(track):
            track.id_in_backend = new_ident
            with open(self.gpx_path(new_ident), 'w', encoding='utf-8') as out_file:
                out_file.write(track.xml())
            self._set_filetime(track)
        return new_ident

    def detach(self):
        """also remove temporary directory."""
        super(Directory, self).detach()
        if self.account.is_temporary:
            remove_directory(self.url)

    @classmethod
    def _check_id_legal(cls, value):
        """Check if value is a legal id.

        If not, raise ValueError.

        """
        # it is not necessary to call BackendBase._check_id_legal
        if value is not None:
            if '/' in value:
                raise ValueError('/ not allowed as id_in_backend for Directory')
            if value == '.':
                raise ValueError('. not allowed as id_in_backend for Directory')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""This module defines :class:`~gpxity.diff.BackendDiff`."""

# pylint: disable=protected-access

import datetime

from collections import defaultdict
from difflib import SequenceMatcher

from .backend import Backend
from .track import Track

__all__ = ['BackendDiff']


class BackendDiff:

    """Compares two backends.directory.

    Args:
        left (Backend): A backend, a track or a list of either
        right (Backend): Same as for the left side

    Attributes:
        left(:class:`BackendDiffSide`): Attributes for the left side
        right(:class:`BackendDiffSide`): Attributes for the right side
        identical(list(Track)): Tracks appearing on both sides.
        similar(list(Pair)): Pairs of Tracks are on both sides with
            differences. This includes all tracks having at least
            100 identical positions without being identical.
        diff_flags: T=time, D=description, C=category, S=status,
            K=keywords, P=positions, Z=time offset

    """

    # pylint: disable=too-few-public-methods

    diff_flags = 'TDCSKPZ'

    class Pair:
        """Holds two comparable Items and the diff result
        Attributes:
            differences(dict()): Keys are Flags for differences, see BackendDiff.diff_flags.
                    Values is a list(str) with additional info
        """
        def __init__(self, left, right):
            """See class docstring."""
            self.left = left
            self.right = right
            self.differences = self.__compare()

        def __compare_metadata(self):
            """Compare some metadata between left and right.

            Returns:
                a dict with the differences

            """
            result = defaultdict(list)

            def compare_attribute(code, attribute, func=None):
                """Compare a specific attribute."""
                left_value = getattr(self.left, attribute) or ''
                right_value = getattr(self.right, attribute) or ''
                if func:
                    left_value = func(left_value)
                    right_value = func(right_value)
                if left_value != right_value:
                    result[code].append('"{}" <> "{}"'.format(left_value, right_value))

            compare_attribute('T', 'title')
            compare_attribute('D', 'description')
            compare_attribute('C', 'category')
            compare_attribute('K', 'keywords', lambda x: ', '.join(x))  # pylint: disable=unnecessary-lambda
            compare_attribute('S', 'public', lambda x: 'public' if x else 'private')
            return result

        def __compare(self):  # noqa
            """Compare both tracks.

            Returns:
                defaultdict(list): Keys are Flags for differences, see BackendDiff.diff_flags.
                    Values is a list(str) with additional info

            """
            # pylint: disable=too-many-locals, too-many-branches, too-many-nested-blocks

            result = self.__compare_metadata()

            def lists(track):
                """Returns two lists of tuples: once with time, once without time."""
                times = list()
                positions = list()
                for _ in track.points():
                    times.append(_.first_time or datetime.datetime(year=1970, month=1, day=1))
                    positions.append(tuple([_.latitude or 0, _.longitude or 0, _.elevation or 0]))  # noqa
                return times, positions

            def pretty_times(time1, time2):
                """If time2 has the same date, use only the time."""
                if time1.date() == time2.date():
                    time2 = time2.first_time()
                return time1, time2

            left_times, left_positions = lists(self.left)
            right_times, right_positions = lists(self.right)
            for tag, left_start, left_end, right_start, right_end in SequenceMatcher(
                    None, left_positions, right_positions).get_opcodes():
                left_found = left_positions[left_start:left_end]
                right_found = right_positions[right_start:right_end]
                for idx, _ in enumerate(left_found):
                    left_found[idx] = list(_)
                    left_found[idx].append(left_times[left_start + idx])
                for idx, _ in enumerate(right_found):
                    right_found[idx] = list(_)
                    right_found[idx].append(right_times[right_start + idx])
                if tag == 'delete':
                    result['P'].append(
                        'points between {} and {} are missing on the right'.format(
                            *pretty_times(left_times[left_start], left_times[left_end - 1])))
                elif tag == 'insert':
                    result['P'].append(
                        'points between {} and {} are missing on the left'.format(
                            *pretty_times(right_times[right_start], right_times[right_end - 1])))
                elif tag == 'replace':
                    if [x[:2] for x in left_found] == [x[:2] for x in right_found]:
                        if len(
                                {(right_found[x][3] - left_found[x][3])
                                 for x in range(len(left_found))}) == 1:
                            time1, time2 = pretty_times(left_found[0][3], left_found[-1][3])
                            timedelta = right_found[0][3] - left_found[0][3]
                            if timedelta:
                                result['Z'].append(
                                    '{} points between {} and {} on the left are {} later on the right'.format(
                                        len(left_found), time1, time2, timedelta))
                            else:
                                # if points are different but times are the same, it must be the height. Ignore that.
                                pass
                        else:
                            result['Z'].append('Points have different times')
                    else:
                        result['P'].append(
                            'points between {} and {} are different'.format(
                                *pretty_times(
                                    min([left_times[left_start], right_times[right_start]]),
                                    max([left_times[left_end - 1], right_times[right_end - 1]]))))
                        for left, right in zip(left_found, right_found):
                            for data, sign in ((left, '<'), (right, '>')):
                                result['P'].append(
                                    '  {sign} {data[0]:8.6f} {data[1]:8.6f} {data[2]:5.2f} {data[3]}'.format(
                                        sign=sign, data=data))

            # some files have a problem with the time zone
            _ = self.left.time_offset(self.right)
            if _:
                result['Z'].append('Time offset: {}'.format(_))

            return result

    class BackendDiffSide:
        """Represents a side (left or right) in BackendDiff.

        Attributes:
            tracks: An Track, a list of tracks, a backend or a list of backends
            exclusive(list): Acivities existing only on this side
        """

        # pylint: disable=too-few-public-methods

        def __init__(self, tracks):
            """See class docstring."""
            self.tracks = list(self.flatten(tracks))
            self.build_positions()
            self.exclusive = []

        @staticmethod
        def flatten(whatever):
            """Flatten Backends or Tracks into a list of tracks."""
            if isinstance(whatever, list):
                for list_item in whatever:
                    if isinstance(list_item, Track):
                        yield list_item
                    elif isinstance(list_item, Backend):
                        for _ in list_item:
                            yield _
            else:
                if isinstance(whatever, Track):
                    yield whatever
                elif isinstance(whatever, Backend):
                    for _ in whatever:
                        yield _

        def build_positions(self):
            """Return a set of long/lat tuples."""
            for _ in self.tracks:
                _.positions = {(x.longitude, x.latitude) for x in _.points()}

        def _find_exclusives(self, matched):
            """use data from the other side."""
            for _ in self.tracks:
                if _ not in matched:
                    self.exclusive.append(_)

    def __init__(self, left, right):
        """See class docstring."""
        self.similar = []
        self.identical = []
        matched = []
        self.left = BackendDiff.BackendDiffSide(left)
        self.right = BackendDiff.BackendDiffSide(right)
        # pylint: disable=too-many-nested-blocks
        for left_track in self.left.tracks:
            for right_track in self.right.tracks:
                if left_track == right_track:
                    self.identical.append(left_track)
                    matched.append(left_track)
                    matched.append(right_track)
                else:
                    maxlen = max(len(left_track.positions), len(right_track.positions))
                    if len(left_track.positions & right_track.positions) >= maxlen * 0.9:
                        self.similar.append(BackendDiff.Pair(left_track, right_track))
                        matched.append(left_track)
                        matched.append(right_track)
        self.left._find_exclusives(matched)
        self.right._find_exclusives(matched)
